#!/usr/bin/env python
# -*- coding: UTF8 -*-
import argparse
import copy
import curses
import json
import math
import os
import socket
import struct
import sys
import time
import traceback
from itertools import cycle

import msgpack
import pygame
from loguru import logger
from pygame import gfxdraw
from pygame.rect import Rect

from pygomas.bditroop import CLASS_NONE, CLASS_SOLDIER, CLASS_MEDIC, CLASS_ENGINEER, CLASS_FIELDOPS
from pygomas.config import TEAM_AXIS, TEAM_ALLIED, TEAM_NONE
from pygomas.pack import PACK_MEDICPACK, PACK_AMMOPACK, PACK_OBJPACK
from .server import (
    MSG_TYPE,
    MSG_BODY,
    TCP_AGL,
    TCP_COM,
    TCP_MAP,
    TCP_TIME,
    TCP_ERR,
    MSG_AGENTS,
    MSG_PACKS,
    MSG_CONTENT_NAME,
    MSG_CONTENT_TYPE,
    MSG_CONTENT_TEAM,
    MSG_CONTENT_HEALTH,
    MSG_CONTENT_AMMO,
    MSG_CONTENT_CARRYINGFLAG,
    MSG_CONTENT_POSITION,
    MSG_CONTENT_HEADING,
    WELCOME_MSG,
    READY_MSG,
    QUIT_MSG,
    ACCEPT_MSG,
)

draw_rect = False


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i: i + n]


def get_angle(angx, angy):
    if angx == 0:
        div = 1000
    else:
        div = angy / angx
    if angy >= 0 and angx >= 0:  # q1
        angle = math.atan(div) * (180 / math.pi)
    elif angy >= 0 >= angx:  # q2
        angle = math.atan(div) * (180 / math.pi) + 180
    elif angy <= 0 and angx <= 0:  # q3
        angle = math.atan(div) * (180 / math.pi) + 180
    else:  # q4
        angle = math.atan(div) * (180 / math.pi) + 360
    return angle


class Render(object):
    def __init__(
            self,
            address="localhost",
            port=8001,
            maps=None,
            text=False,
            replay=False,
            dump=False,
            log="match.log",
            wait_fps=0.033,
    ):
        self.address = address
        self.port = port
        self.maps_path = maps
        self.text = text
        self.s = None
        self.screen = None
        self.font = None

        self.objective_x = -1
        self.objective_y = -1
        self.allied_base = None
        self.axis_base = None
        self.graph = {}
        self.agents = {}
        self.dins = {}
        self.factor = 2

        self.iteration = 0

        self.tile_size = 24
        self.horizontal_tiles = 32
        self.vertical_tiles = 32

        self.map_width = self.tile_size * self.horizontal_tiles
        self.map_height = self.tile_size * self.vertical_tiles

        self.xdesp = 0
        self.ydesp = 0
        self.size = [self.map_width, self.map_height]

        self.show_help = False
        self.show_fovs = True
        self.show_info = True
        self.quit = False

        self.fps = list()

        this_dir, _ = os.path.split(__file__)
        self.assets_path = f"{this_dir}{os.sep}assets{os.sep}"
        self.allied_base_sprite = pygame.image.load(
            os.path.join(self.assets_path, "base4.png")
        )
        self.axis_base_sprite = pygame.image.load(
            os.path.join(self.assets_path, "base3.png")
        )
        self.wall_sprite = pygame.image.load(
            os.path.join(self.assets_path, "wall2.png")
        )
        self.terrain_sprites = cycle([
            pygame.image.load(os.path.join(self.assets_path, "terrain.jpg")),
            pygame.image.load(os.path.join(self.assets_path, "grass2.jpg")),
            pygame.image.load(os.path.join(self.assets_path, "grass.jpg"))]
        )
        self.terrain_sprite = next(self.terrain_sprites)

        self.pack_images = {
            PACK_MEDICPACK: pygame.image.load(os.path.join(self.assets_path, "medpack2.png")),
            PACK_AMMOPACK: pygame.image.load(os.path.join(self.assets_path, "ammopack.png"))
        }

        self.sprites = {}
        self.graves = {}
        self.flag_sprite = None

        self.replay = replay
        self.dump = dump
        self.log = log
        self.file = None
        self.game_log = []
        self.wait_fps = wait_fps

    def main(self):
        if self.text:
            # curses.wrapper(self._main)
            self._main()
        else:
            self._main()

    def read_msg(self):
        if not self.replay:
            try:
                size_of_msg = self.s.recv(4)
            except socket.timeout:
                size_of_msg = 0
            if not size_of_msg:
                return None
            size_of_msg = struct.unpack(">I", size_of_msg)[0]

            data = bytes()
            while len(data) < size_of_msg:
                packet = self.s.recv(size_of_msg - len(data))
                if not packet:
                    break
                data += packet

            if len(data) == 0:
                data = None
            data = msgpack.unpackb(data, raw=False)
            return data
        else:
            if len(self.game_log) > 0:
                data = self.game_log.pop(0)
                if data:
                    data = json.loads(data)
                    data = self._load_json(data)
                    return data
            return {MSG_TYPE: TCP_COM, MSG_BODY: QUIT_MSG}

    def _load_json(self, data):
        if isinstance(data, dict):
            new_data = dict()
            for key, value in data.items():
                new_data[int(key)] = self._load_json(value)
        elif isinstance(data, list):
            new_data = list()
            for element in data:
                new_data.append(self._load_json(element))
        else:
            new_data = data
        return new_data

    def _main(self, stdscr=None):
        error = False
        if not self.dump:
            if not self.text:
                # Init pygame
                pygame.init()
                self.font = pygame.font.SysFont("ttf-font-awesome", 16)
                self.help_font = pygame.font.SysFont("ttf-font-awesome", 42)

                # Set the height and width of the self.screen
                self.screen = pygame.display.set_mode(self.size)

            else:
                stdscr = curses.initscr()
                curses.start_color()
                curses.noecho()
                curses.cbreak()
                stdscr.keypad(1)

        else:
            logger.success(f"Dumping log to {self.log}")
            self.file = open(self.log, "w")

        try:
            if self.replay:
                logger.success(f"Loading log from {self.log}")
                with open(self.log, "r") as rfile:
                    self.game_log = rfile.read()
                self.game_log = self.game_log.split("\nSEP\n")
            else:
                # Init socket
                self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                if self.s:
                    self.s.settimeout(0.1)
                    self.s.connect((self.address, self.port))

            while not self.quit:
                start_time = time.time()
                if not self.dump:
                    self.draw(stdscr)
                if self.replay:
                    time.sleep(self.wait_fps)

                data = self.read_msg()

                if data is None:
                    continue

                if data[MSG_TYPE] == TCP_COM:
                    if data[MSG_BODY] == WELCOME_MSG:
                        if not self.replay:
                            msg_to_send = msgpack.packb(
                                {MSG_TYPE: TCP_COM, MSG_BODY: READY_MSG},
                                use_bin_type=True,
                            )
                            size_of_package = len(msg_to_send)
                            msg = struct.pack(">I", size_of_package) + msg_to_send
                            self.s.send(msg)
                    elif data[MSG_BODY] == ACCEPT_MSG:
                        pass
                    elif data[MSG_BODY] == QUIT_MSG:
                        self.quit = True

                elif data[MSG_TYPE] == TCP_MAP:
                    if self.dump:
                        self.dump_data(data)
                    else:
                        mapname = data[MSG_BODY]
                        result = self.load_map(mapname)
                        if result["status"] != "ok":
                            error = result["value"]
                            self.quit = True

                        if not self.text:
                            self.init_sprites()

                elif data[MSG_TYPE] == TCP_AGL:
                    if self.dump:
                        self.dump_data(data)
                    else:
                        self.agl_parse(data[MSG_BODY])
                        self.fps.append(1 / (time.time() - start_time))

                elif data[MSG_TYPE] == TCP_TIME:
                    pass
                elif data[MSG_TYPE] == TCP_ERR:
                    pass
                else:
                    # Unknown message type
                    pass

        except Exception as e:
            tb = traceback.format_exc()
            error = str(e) + "\n" + tb
            if self.text:
                curses.nocbreak()
                # self.stdscr.keypad(False)
                curses.echo()
                curses.endwin()
            logger.exception("Exception.")

        finally:
            if not self.replay:
                logger.info("Sending QUIT message...")
                msg_to_send = msgpack.packb(
                    {MSG_TYPE: TCP_COM, MSG_BODY: QUIT_MSG}, use_bin_type=True,
                )
                size_of_package = len(msg_to_send)
                msg = struct.pack(">I", size_of_package) + msg_to_send
                self.s.send(msg)

                logger.info("Closing...")
                # Close socket
                self.s.close()

            if not self.dump:
                if not self.text:
                    pygame.quit()
                else:
                    curses.nocbreak()
                    # self.stdscr.keypad(False)
                    curses.echo()
                    curses.endwin()
            else:
                logger.info("Closing...")
                self.file.close()
            if error:
                logger.error(str(error))

    def dump_data(self, data):
        self.file.write(json.dumps(data))
        self.file.write("\nSEP\n")

    def agl_parse(self, data):
        self.dins = {}

        for agent in data[MSG_AGENTS]:
            self.agents[agent[MSG_CONTENT_NAME]] = agent

        for pack in data[MSG_PACKS]:
            self.dins[pack[MSG_CONTENT_NAME]] = pack

    def draw(self, stdscr):
        if self.text:
            self.textdraw(stdscr)
        else:
            self.pygamedraw()

    def pygamedraw(self):
        global draw_rect

        events = pygame.event.get()
        for event in events:
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_LEFT:
                    self.xdesp += self.tile_size
                elif event.key == pygame.K_RIGHT:
                    self.xdesp -= self.tile_size
                elif event.key == pygame.K_DOWN:
                    self.ydesp -= self.tile_size
                elif event.key == pygame.K_UP:
                    self.ydesp += self.tile_size
                elif event.key == pygame.K_x:
                    self.tile_size += 2
                elif event.key == pygame.K_z:
                    self.tile_size -= 2
                elif event.key == pygame.K_f:
                    self.show_fovs = not self.show_fovs
                elif event.key == pygame.K_i:
                    self.show_info = not self.show_info
                elif event.key == pygame.K_b:
                    self.terrain_sprite = next(self.terrain_sprites)
                elif event.key == pygame.K_r:
                    draw_rect = not draw_rect
                elif event.key == pygame.K_h:
                    self.show_help = not self.show_help
                elif event.key == pygame.K_ESCAPE:
                    self.quit = True
            elif event.type == pygame.QUIT:
                self.quit = True

        # Clear screen
        self.pygame_draw_background()

        # Draw bases
        self.pygame_draw_bases()

        # Draw Map
        self.pygame_draw_walls()

        # Draw FPS
        self.pygame_draw_fps()

        # Draw items
        self.pygame_draw_packs()

        # Draw units
        self.pygame_draw_units()

        # Show help
        self.pygame_draw_help()

        pygame.display.update()  # flip()
        self.iteration += 1

    def pygame_draw_units(self):
        for name, agent in self.agents.items():
            health = float(agent[MSG_CONTENT_HEALTH])
            posx = (
                    int(float(agent[MSG_CONTENT_POSITION][0]) * self.tile_size / 8.0)
                    + self.xdesp
            )
            posy = (
                    int(float(agent[MSG_CONTENT_POSITION][2]) * self.tile_size / 8.0)
                    + self.ydesp
            )

            if float(health) > 0:
                carrying = agent[MSG_CONTENT_CARRYINGFLAG]

                # agent_type = {0: "X", 1: "*", 2: "+", 3: "Y", 4: "^"}.get(
                agent_type = {CLASS_NONE: "soldier", CLASS_SOLDIER: "soldier", CLASS_MEDIC: "medic", CLASS_ENGINEER: "engineer",
                              CLASS_FIELDOPS: "fieldops"}.get(agent[MSG_CONTENT_TYPE], "soldier")

                team = {TEAM_ALLIED: (255, 50, 50), TEAM_AXIS: (100, 100, 255)}.get(
                    agent[MSG_CONTENT_TEAM], (255, 255, 0)
                )

                team_aplha = {TEAM_ALLIED: (255, 100, 100, 100), TEAM_AXIS: (100, 100, 255, 100)}.get(
                    agent[MSG_CONTENT_TEAM], (255, 255, 0, 255)
                )

                ammo = float(agent[MSG_CONTENT_AMMO])

                # print avatar
                if name not in self.sprites:
                    sprite = MySprite(name=agent_type, team=agent[MSG_CONTENT_TEAM], num_sprites=8, scale=0.55)
                    self.sprites[name] = sprite
                angx = float(agent[MSG_CONTENT_HEADING][0])
                angy = float(agent[MSG_CONTENT_HEADING][2])
                self.sprites[name].update(posx, posy, angx, angy)
                self.sprites[name].draw(self.screen)

                if self.show_info:
                    # print name
                    text = self.font.render(agent[MSG_CONTENT_NAME], True, (255, 255, 255))
                    self.screen.blit(
                        text,
                        (
                            posx - text.get_width() // 2 + 25,
                            posy - text.get_height() // 2 - 25,
                        ),
                    )
                    rect = self.sprites[name].rect
                    health_bar = Rect(rect.bottomleft[0], rect.bottomleft[1], rect.width, 4)
                    pygame.draw.rect(self.screen, (255, 0, 0), health_bar)
                    health_bar = Rect(rect.bottomleft[0], rect.bottomleft[1], min(100, health * rect.width / 100), 4)
                    pygame.draw.rect(self.screen, (0, 255, 0), health_bar)

                    ammo_bar = Rect(rect.bottomleft[0], rect.bottomleft[1] + 5, rect.width, 4)
                    pygame.draw.rect(self.screen, (255, 255, 255), ammo_bar)
                    ammo_bar = Rect(rect.bottomleft[0], rect.bottomleft[1] + 5, min(100, ammo * rect.width / 100), 4)
                    pygame.draw.rect(self.screen, (106, 115, 120), ammo_bar)

                # is carring flag?
                if carrying:
                    flagx, flagy = self.sprites[name].rect.topright
                    self.flag_sprite.update(flagx, flagy, 0, -1, force_animation=True)
                    self.flag_sprite.draw(self.screen)

                # print fov
                if self.show_fovs:
                    # compute direction
                    angx = float(agent[MSG_CONTENT_HEADING][0])
                    angy = float(agent[MSG_CONTENT_HEADING][2])

                    angle = get_angle(angx, angy)

                    for j in range(0, int(48 * (self.tile_size / 8)), 1):
                        pygame.gfxdraw.arc(
                            self.screen,
                            posx,
                            posy,
                            j,
                            int(-45 + angle),
                            int(45 + angle),
                            team_aplha,
                        )

            else:
                team = agent[MSG_CONTENT_TEAM]
                if team not in self.graves:
                    sprite = MySprite(name="grave", team=team, num_sprites=1, scale=0.3)
                    self.graves[team] = sprite
                self.graves[team].update(posx, posy, 0, -1)
                self.graves[team].draw(self.screen)

    def pygame_draw_packs(self):
        for pack in self.dins.values():
            posx = int(
                pack[MSG_CONTENT_POSITION][0] * (self.tile_size / 8.0) + self.xdesp
            )
            posy = int(
                pack[MSG_CONTENT_POSITION][2] * (self.tile_size / 8.0) + self.ydesp
            )

            item_type = pack[MSG_CONTENT_TYPE]  # {1001: "M", 1002: "A", 1003: "F"}.get(
            # pack[MSG_CONTENT_TYPE], "X"
            # )

            if item_type != PACK_OBJPACK:

                '''color = {
                    1001: (88, 214, 141),  # (255, 255, 255),
                    1002: (155, 89, 182),  # (255, 255, 255),
                    1003: (255, 255, 0),
                }.get(pack[MSG_CONTENT_TYPE], "X")

                pygame.draw.circle(self.screen, color, [posx, posy], 6)
                text = self.font.render(item_type, True, (0, 0, 0))
                self.screen.blit(
                    text, (posx - text.get_width() // 2, posy - text.get_height() // 2)
                )'''
                self.screen.blit(
                    self.pack_images[item_type],
                    (
                        posx, posy
                    ),
                )
            else:
                self.flag_sprite.update(posx, posy, 0, -1, force_animation=True)
                self.flag_sprite.draw(self.screen)

    def pygame_draw_fps(self):
        if len(self.fps) > 5:
            fps = int(sum(self.fps[-5:]) / 5)
            text = self.font.render(str(fps) + " FPS", True, (255, 255, 255))
            self.screen.blit(text, (1, 1))

    def pygame_draw_walls(self):
        color_wall = (100, 100, 100)
        for y in range(0, len(list(self.graph.items()))):
            for x in range(0, 32):
                try:
                    if list(self.graph.items())[y][1][x] == "*":
                        """pygame.draw.rect(
                            self.screen,
                            color_wall,
                            (
                                x * self.tile_size + self.xdesp,
                                y * self.tile_size + self.ydesp,
                                self.tile_size,
                                self.tile_size,
                            ),
                        )"""
                        self.screen.blit(
                            self.wall_sprite,
                            (
                                x * self.tile_size + self.xdesp,
                                y * self.tile_size + self.ydesp,
                            ),
                        )
                except:
                    pass

    def pygame_draw_bases(self):
        if self.allied_base is not None:
            if len(self.allied_base) == 5:
                draw = bool(int(self.allied_base[4]))
            else:
                draw = True
            if draw:
                color = (255, 0, 0)
                xpos = int(self.allied_base[0]) * self.tile_size + self.xdesp
                ypos = int(self.allied_base[1]) * self.tile_size + self.ydesp
                xwidth = (
                        int(self.allied_base[2]) * self.tile_size
                        - xpos
                        + self.tile_size
                        + self.xdesp
                )
                ywidth = (
                        int(self.allied_base[3]) * self.tile_size
                        - ypos
                        + self.tile_size
                        + self.ydesp
                )

                self.screen.blit(self.allied_base_sprite, (xpos, ypos))
        if self.axis_base is not None:
            if len(self.axis_base) == 5:
                draw = bool(int(self.axis_base[4]))
            else:
                draw = True
            if draw:
                color = (0, 0, 255)
                xpos = int(self.axis_base[0]) * self.tile_size + self.xdesp
                ypos = int(self.axis_base[1]) * self.tile_size + self.ydesp
                xwidth = (
                        int(self.axis_base[2]) * self.tile_size
                        - xpos
                        + self.tile_size
                        + self.xdesp
                )
                ywidth = (
                        int(self.axis_base[3]) * self.tile_size
                        - ypos
                        + self.tile_size
                        + self.ydesp
                )

                self.screen.blit(self.axis_base_sprite, (xpos, ypos))

    def pygame_draw_background(self):
        color_background = (189, 236, 182)  # (245, 245, 220)
        # pygame.draw.rect(
        #    self.screen, color_background, (0, 0, self.map_width, self.map_height)
        # )
        self.screen.blit(self.terrain_sprite, (0, 0))

    def textdraw(self, stdscr):
        height, width = stdscr.getmaxyx()

        # Draw Map
        for name, v in list(self.graph.items()):
            try:
                newline = ""
                for char in v:
                    newline += char * self.factor
                if height > name:
                    stdscr.addstr(name, 0, str(newline))
            except:
                pass

        # Draw bases
        # ALLIED BASE
        if self.allied_base is not None:
            curses.init_pair(4, curses.COLOR_WHITE, curses.COLOR_RED)  # ALLIED BASE
            for y in range(int(self.allied_base[1]), int(self.allied_base[3])):
                for x in range(
                        int(self.allied_base[0]) * self.factor,
                        int(self.allied_base[2]) * self.factor,
                ):
                    if height > y:
                        stdscr.addstr(y, x, " ", curses.color_pair(4))

        # AXIS BASE
        if self.axis_base is not None:
            curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLUE)  # AXIS BASE
            for y in range(int(self.axis_base[1]), int(self.axis_base[3])):
                for x in range(
                        int(self.axis_base[0]) * self.factor,
                        int(self.axis_base[2]) * self.factor,
                ):
                    if height > y:
                        stdscr.addstr(y, x, " ", curses.color_pair(3))

        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_YELLOW)
        # PACKS
        # for k, v in list(self.dins.items()):
        for pack in self.dins.values():
            #  Type
            symbol = {"1001": "M", "1002": "A", "1003": "F"}.get(
                str(pack[MSG_CONTENT_TYPE]), "X"
            )

            y = int(float(pack[MSG_CONTENT_POSITION][2]) / 8)
            x = int(float(pack[MSG_CONTENT_POSITION][0]) / (8 / self.factor))
            if height > y:
                stdscr.addstr(y, x, symbol, curses.color_pair(2))

        curses.init_pair(5, curses.COLOR_BLACK, curses.COLOR_RED)  # ALLIED
        curses.init_pair(6, curses.COLOR_WHITE, curses.COLOR_BLUE)  # AXIS
        curses.init_pair(7, curses.COLOR_BLACK, curses.COLOR_WHITE)  #  OTHER / DEAD

        # AGENTS
        stats_allied = []  # ""
        stats_axis = []  # ""
        # for k, v in list(self.agents.items()):
        for agent in self.agents.values():
            name = agent[MSG_CONTENT_NAME]
            name = name[:5] + ".." + name[-5:] if len(name) > 12 else name
            # Type
            symbol = {1: "*", 2: "+", 3: "Y", 4: "^"}.get(agent[MSG_CONTENT_TYPE], "X")

            # Team (or Carrier)
            team_color = {100: 5, 200: 6, }.get(agent[MSG_CONTENT_TEAM], 1)

            if agent[MSG_CONTENT_CARRYINGFLAG]:
                team_color = 2

            # Draw in map
            y = int(float(agent[MSG_CONTENT_POSITION][2]) / 8)
            x = int(float(agent[MSG_CONTENT_POSITION][0]) / (8 / self.factor))
            health = int(agent[MSG_CONTENT_HEALTH])
            ammo = int(agent[MSG_CONTENT_AMMO])
            if health > 0:
                if height > y:
                    stdscr.addstr(y, x, symbol, curses.color_pair(team_color))  #  Alive
            else:
                if height > y:
                    stdscr.addstr(y, x, "D", curses.color_pair(7))  #  Dead
            # Write stats
            if agent[MSG_CONTENT_TEAM] == 100:
                if health > 0:
                    stats_allied.append(
                        f" | {symbol} {name.ljust(4)} {health:03d} {ammo:03d} "
                    )
                else:
                    stats_allied.append(f" | {symbol} {name.ljust(4)} --- --- ")
            elif agent[MSG_CONTENT_TEAM] == 200:
                if health > 0:
                    # stats_axis += " | %s %s %03d %03d " % (c, k, int(v["health"]), int(v["ammo"]))
                    stats_axis.append(
                        f" | {symbol} {name.ljust(4)} {health:03d} {ammo:03d} "
                    )
                else:
                    stats_axis.append(f" | {symbol} {name.ljust(4)} --- --- ")

        row = 33
        for _agents in chunks(stats_allied, 4):
            line = "".join(_agents)
            if height > row:
                stdscr.addstr(row, 1, str(line), curses.color_pair(5))
            row += 1
        # stdscr.addstr(34, 1, blank)
        for _agents in chunks(stats_axis, 4):
            line = "".join(_agents)
            if height > row:
                stdscr.addstr(row, 1, str(line), curses.color_pair(6))
            row += 1

        # Draw FPS
        if len(self.fps) > 5:
            fps = int(sum(self.fps[-5:]) / 5)
            if height > row:
                stdscr.addstr(row, 1, "{:03d} FPS".format(fps), curses.color_pair(7))
            row += 1

        # Refresh screen
        try:
            stdscr.refresh()
        except:
            pass

    def load_map(self, map_name):
        try:
            if self.maps_path is not None:
                path = f"{self.maps_path}{os.sep}{map_name}{os.sep}"
            else:
                this_dir, _ = os.path.split(__file__)
                path = f"{this_dir}{os.sep}maps{os.sep}{map_name}{os.sep}"

            if os.path.exists(f"{path}{map_name}.json"):
                with open(f"{path}{map_name}.json") as f:
                    mapf = json.load(f)
                    self.objective_x = int(mapf["objective"][0])
                    self.objective_y = int(mapf["objective"][1])
                    self.allied_base = mapf["spawn"]["allied"]
                    self.axis_base = mapf["spawn"]["axis"]
                    cost = open(f"{path}{mapf['cost_map']['file']}", "r")

            else:
                mapf = open(f"{path}{map_name}.txt", "r")
                cost = open(f"{path}{map_name}_cost.txt", "r")
                for line in mapf.readlines():
                    if "pGomas_OBJECTIVE" in line:
                        splitted_line = line.split()
                        self.objective_x = copy.copy(int(splitted_line[1]))
                        self.objective_y = copy.copy(int(splitted_line[2]))
                    elif "pGomas_SPAWN_ALLIED" in line:
                        splitted_line = line.split()
                        splitted_line.pop(0)
                        self.allied_base = copy.copy(splitted_line)
                    elif "pGomas_SPAWN_AXIS" in line:
                        splitted_line = line.split()
                        splitted_line.pop(0)
                        self.axis_base = copy.copy(splitted_line)
                mapf.close()

            y = 0
            for line in cost.readlines():
                self.graph[y] = line.strip("\r\n")
                y += 1
            cost.close()
            if not self.text:
                logger.info("Map loaded.")
            return {"status": "ok"}
        except Exception as e:
            tb = traceback.extract_stack()
            return {
                "status": "error",
                "value": str(e) + "\n" + "\n".join([repr(i) for i in tb]),
            }

    def init_sprites(self):
        xpos = int(self.allied_base[0]) * self.tile_size + self.xdesp
        ypos = int(self.allied_base[1]) * self.tile_size + self.ydesp
        xwidth = (
                int(self.allied_base[2]) * self.tile_size
                - xpos
                + self.tile_size
                + self.xdesp
        )
        ywidth = (
                int(self.allied_base[3]) * self.tile_size
                - ypos
                + self.tile_size
                + self.ydesp
        )

        self.allied_base_sprite.convert()
        self.allied_base_sprite = pygame.transform.scale(
            self.allied_base_sprite, (xwidth, ywidth)
        )

        xpos = int(self.axis_base[0]) * self.tile_size + self.xdesp
        ypos = int(self.axis_base[1]) * self.tile_size + self.ydesp
        xwidth = (
                int(self.axis_base[2]) * self.tile_size - xpos + self.tile_size + self.xdesp
        )
        ywidth = (
                int(self.axis_base[3]) * self.tile_size - ypos + self.tile_size + self.ydesp
        )

        self.axis_base_sprite.convert()
        self.axis_base_sprite = pygame.transform.scale(
            self.axis_base_sprite, (xwidth, ywidth)
        )

        self.wall_sprite.convert()
        self.wall_sprite = pygame.transform.scale(
            self.wall_sprite, (self.tile_size, self.tile_size)
        )

        self.terrain_sprite.convert()
        self.terrain_sprite = pygame.transform.scale(
            self.terrain_sprite, (self.map_width, self.map_height)
        )

        self.flag_sprite = MySprite(name="flag", team=TEAM_NONE, num_sprites=8, scale=0.1, fr=4)

        scale = 0.15
        self.pack_images = {k: pygame.transform.scale(img, (int(img.get_size()[0] * scale), int(img.get_size()[1] * scale))) for
                            k, img in self.pack_images.items()}

    def pygame_draw_help(self):
        if self.show_help:
            text = self.help_font.render("Show/Hide Field of View: F", True, (255, 255, 255))
            self.screen.blit(text, (150, 70), )

            text = self.help_font.render("Show/Hide Info (Bars and names): I", True, (255, 255, 255))
            self.screen.blit(text, (150, 100), )

            text = self.help_font.render("Change background : B", True, (255, 255, 255))
            self.screen.blit(text, (150, 130), )

            text = self.help_font.render("Show/Hide Bounding Boxes: R", True, (255, 255, 255))
            self.screen.blit(text, (150, 160), )

            text = self.help_font.render("Move camera: Arrows", True, (255, 255, 255))
            self.screen.blit(text, (150, 190), )

            text = self.help_font.render("Zoom In/Out: X,Z", True, (255, 255, 255))
            self.screen.blit(text, (150, 220), )

            text = self.help_font.render("Quit: ESCAPE", True, (255, 255, 255))
            self.screen.blit(text, (150, 250), )


class MySprite(pygame.sprite.Sprite):
    def __init__(self, name: str, team: int, num_sprites: int, folder: str = "assets", scale: float = 1.0, fr=1):
        super(MySprite, self).__init__()
        # adding all the images to sprite array
        this_dir, _ = os.path.split(__file__)

        team = {TEAM_ALLIED: "allied", TEAM_AXIS: "axis", TEAM_NONE: ""}.get(team)
        self.images = [pygame.image.load(f'{this_dir}{os.sep}{folder}{os.sep}{name}_{team}{i}.png') for i in
                       range(1, num_sprites + 1)]
        self.images = [pygame.transform.scale(img, (int(img.get_size()[0] * scale), int(img.get_size()[1] * scale)))
                       for img in self.images]
        self.images = [pygame.transform.rotate(img, -90) for img in self.images]
        self.original_images = [copy.copy(img) for img in self.images]

        # index value to get the image from the array
        # initially it is 0
        self.index = 0

        # now the image that we will display will be the index from the image array
        self.image = self.images[self.index]
        self.x = 0
        self.y = 0

        self.rect = self.image.get_rect(center=(self.x, self.y))
        self.fr = cycle(range(fr))

    def update(self, x, y, angx, angy, force_animation=False):

        fr = next(self.fr)
        if fr == 0:
            if self.x != x or self.y != y or force_animation:
                # when the update method is called, we will increment the index
                self.index += 1

                # if the index is larger than the total images
                if self.index >= len(self.images):
                    # we will make the index to 0 again
                    self.index = 0

        self.x, self.y = x, y

        angle = get_angle(angx, angy)

        self.image = pygame.transform.rotate(self.original_images[self.index], -angle)
        self.rect = self.image.get_rect(center=(self.x, self.y))

    def draw(self, surface):
        if draw_rect:
            pygame.draw.rect(surface, (255, 0, 0), self.rect, 1)
        surface.blit(self.image, self.rect)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ip", default="localhost", help="Manager's address to connect the render"
    )
    parser.add_argument(
        "--port", default=8001, help="Manager's port to connect the render"
    )
    parser.add_argument(
        "--maps", default=None, help="The path to your custom maps directory"
    )
    parser.add_argument("--text", default=False, help="Use text render")

    parser.add_argument("--log", default="match.log", help="File to save the game")

    args = parser.parse_args()
    render = Render(args.ip, args.port, args.maps, args.text)
    render.main()
    sys.exit(0)



import unittest
from jgomas.CTerrainMap import CBase, CTerrain, CTerrainMap
from jgomas.CConfig import CConfig


class TestCBase(unittest.TestCase):

    base = CBase()

    def test_getters(self):

        init_x = self.base.GetInitX()
        init_y = self.base.GetInitY()
        init_z = self.base.GetInitZ()
        end_x = self.base.GetEndX()
        end_y = self.base.GetEndY()
        end_z = self.base.GetEndZ()

        self.assertEqual(
            init_x,
            0
        )

        self.assertEqual(
            init_y,
            0
        )

        self.assertEqual(
            init_z,
            0
        )

        self.assertEqual(
            end_x,
            0
        )

        self.assertEqual(
            end_y,
            0
        )

        self.assertEqual(
            end_z,
            0
        )


class TestCTerrain(unittest.TestCase):

    def test_repr(self):
        terrain = CTerrain()

        self.assertEqual(
            terrain.__repr__(),
            "*"
        )

        terrain.m_bCanWalk = True

        self.assertEqual(
            terrain.__repr__(),
            "_"
        )


class TestCTerrainMap(unittest.TestCase):

    def test_getters(self):
        terrain_map = CTerrainMap()

        size_x = terrain_map.GetSizeX()
        size_z = terrain_map.GetSizeZ()
        target_x = terrain_map.GetTargetX()
        target_y = terrain_map.GetTargetY()
        target_z = terrain_map.GetTargetZ()

        self.assertEqual(
            size_x,
            0
        )

        self.assertEqual(
            size_z,
            0
        )

        self.assertEqual(
            target_x,
            0
        )

        self.assertEqual(
            target_y,
            0
        )

        self.assertEqual(
            target_z,
            0
        )

    def test_can_walk(self):

        terrain_map = CTerrainMap()
        x = 0
        z = 0

        can_walk = terrain_map.CanWalk(x, z)

        self.assertFalse(
            can_walk
        )

        terrain_map.m_iSizeX = 1
        terrain_map.m_iSizeZ = 1
        terrain_map.m_Terrain = []
        terrain_map.m_Terrain.append([])
        terrain = CTerrain()
        terrain.m_bCanWalk = True
        terrain_map.m_Terrain[0].append(terrain)

        can_walk = terrain_map.CanWalk(x, z)

        self.assertTrue(
            can_walk
        )

    def test_get_cost(self):

        terrain_map = CTerrainMap()
        x = 0
        z = 0

        cost = terrain_map.GetCost(x, z)

        self.assertEqual(
            cost,
            20000
        )

        terrain_map.m_iSizeX = 1
        terrain_map.m_iSizeZ = 1
        terrain_map.m_Terrain = []
        terrain_map.m_Terrain.append([])
        terrain = CTerrain()
        terrain.m_iCost = 10
        terrain_map.m_Terrain[0].append(terrain)

        cost = terrain_map.GetCost(x, z)

        self.assertEqual(
            cost,
            10
        )

    def test_load_map(self):

        terrain_map = CTerrainMap()
        config = CConfig()
        config.setDataPath("./tests/test_maps")
        file = "map_01"

        terrain_map.LoadMap(file, config)

        self.assertGreater(
            len(terrain_map.m_Terrain),
            0
        )

        self.assertGreater(
            len(terrain_map.m_Terrain[0]),
            0
        )

    def test_load_map_failed(self):

        terrain_map = CTerrainMap()
        config = CConfig()
        config.setDataPath("./tests/test_maps")
        file = "map_00"

        terrain_map.LoadMap(file, config)

        self.assertEqual(
            terrain_map.m_iSizeX,
            0
        )

        self.assertEqual(
            terrain_map.m_iSizeZ,
            0
        )

    def test_str(self):
        terrain_map = CTerrainMap()
        config = CConfig()
        config.setDataPath("./tests/test_maps")
        file = "map_01"

        terrain_map.LoadMap(file, config)
        res = str(terrain_map)

        self.assertIsInstance(
            res,
            str
        )



import random

from loguru import logger
from spade.behaviour import OneShotBehaviour

from .config import POWER_UNIT
from .bditroop import BDITroop, CLASS_MEDIC
from .ontology import MEDIC_SERVICE
from .medicpack import MedicPack


class BDIMedic(BDITroop):
    packs_delivered = 0
    medic_pack_offset = 5

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.services.append(MEDIC_SERVICE)
        self.eclass = CLASS_MEDIC

    def add_custom_actions(self, actions):
        super().add_custom_actions(actions)

        @actions.add(".cure", 0)
        def _cure(agent, term, intention):
            class CreateMedicPackBehaviour(OneShotBehaviour):
                async def run(self):
                    await self.agent.create_medic_pack()

            b = CreateMedicPackBehaviour()
            self.add_behaviour(b)
            yield

    def perform_medic_action(self):
        # We can give medic paks if we have power enough...
        if self.get_power() >= POWER_UNIT:
            self.use_power()
            return True

        return False

    async def create_medic_pack(self):
        """
        Creates medic packs if possible.

        This method allows to create medic packs if there is enough power in the agent's power bar.

        :returns number of medic packs created
        """
        logger.info("{} Creating medic packs.".format(self.name))
        while self.perform_medic_action():
            BDIMedic.packs_delivered += 1
            name = "medicpack_{}_{}@{}".format(
                self.jid.localpart, BDIMedic.packs_delivered, self.jid.domain
            )
            x = self.movement.position.x + random.random() * BDIMedic.medic_pack_offset
            z = self.movement.position.z + random.random() * BDIMedic.medic_pack_offset

            while not self.check_static_position(x, z):
                x = (
                    self.movement.position.x
                    + random.random() * BDIMedic.medic_pack_offset
                )
                z = (
                    self.movement.position.z
                    + random.random() * BDIMedic.medic_pack_offset
                )

            try:
                pack = MedicPack(
                    name=name, passwd="secret", x=x, z=z, manager_jid=self.manager
                )
                await pack.start()
            except Exception as e:
                logger.warning(
                    "Medic {} could not create MedicPack: {}".format(self.name, e)
                )



import unittest
import math
from jgomas.CMobile import CMobile
from jgomas.Vector3D import Vector3D


class TestCMobile(unittest.TestCase):

    def test_set_size(self):
        max_x = 5
        max_z = 10
        mobile = CMobile()
        mobile.SetSize(max_x, max_z)

        self.assertEqual(
            mobile.m_MinX,
            8
        )

        self.assertEqual(
            mobile.m_MinZ,
            8
        )

        self.assertEqual(
            mobile.m_MaxX,
            (max_x * 8) - 8
        )

        self.assertEqual(
            mobile.m_MaxZ,
            (max_z * 8) - 8
        )

    def test_calculate_position(self):
        dt = 0.001
        mobile = CMobile()
        mobile.m_Velocity = Vector3D(x=10, y=20, z=30)
        mobile.CalculatePosition(dt)

        self.assertEqual(
            mobile.m_Position.x,
            mobile.m_Velocity.x * dt
        )

        self.assertEqual(
            mobile.m_Position.y,
            0
        )

        self.assertEqual(
            mobile.m_Position.z,
            mobile.m_Velocity.z * dt
        )

    def test_calculate_new_orientation(self):
        mobile = CMobile()
        x = 10
        y = 20
        z = 30
        mobile.m_Destination = Vector3D(x=x, y=y, z=z)
        mobile.CalculateNewOrientation()
        leng = math.sqrt((x * x + y * y + z * z))

        self.assertEqual(
            mobile.m_Heading.x,
            x / leng
        )

        self.assertEqual(
            mobile.m_Heading.y,
            y / leng
        )

        self.assertEqual(
            mobile.m_Heading.z,
            z / leng
        )

    def test_calculate_new_destination(self):
        mobile = CMobile()
        radio_x = 10
        radio_y = 20
        max_x = 5
        max_z = 10
        mobile.SetSize(max_x, max_z)
        mobile.CalculateNewDestination(radio_x, radio_y)

        self.assertNotEqual(
            mobile.m_Destination.x,
            0
        )

        self.assertEqual(
            mobile.m_Destination.y,
            0
        )

        self.assertNotEqual(
            mobile.m_Destination.z,
            0
        )

    def test_getters(self):
        mobile = CMobile()

        destination = mobile.getDestination()
        position = mobile.getPosition()
        angle = mobile.getdAngle()
        radius = mobile.getdViewRadius()
        heading = mobile.getHeading()
        velocity = mobile.getVelocity()

        self.assertIsInstance(
            destination,
            Vector3D
        )

        self.assertIsInstance(
            position,
            Vector3D
        )

        self.assertEqual(
            angle,
            1
        )

        self.assertEqual(
            radius,
            50
        )

        self.assertIsInstance(
            heading,
            Vector3D
        )

        self.assertIsInstance(
            velocity,
            Vector3D
        )

    def test_setters(self):
        mobile = CMobile()
        vector = Vector3D(x=1, y=2, z=3)
        mobile.setDestination(vector)

        self.assertEqual(
            mobile.m_Destination.x,
            vector.x
        )

        self.assertEqual(
            mobile.m_Destination.y,
            vector.y
        )

        self.assertEqual(
            mobile.m_Destination.z,
            vector.z
        )



import unittest
from jgomas.CSight import CSight
from jgomas.Vector3D import Vector3D


class TestCSight(unittest.TestCase):

    sight = CSight()

    def test_getters(self):

        angle = self.sight.getAngle()
        distance = self.sight.getDistance()
        team = self.sight.getTeam()
        type = self.sight.getType()
        health = self.sight.getHealth()
        position = self.sight.getPosition()

        self.assertEqual(
            angle,
            0
        )

        self.assertEqual(
            distance,
            0
        )

        self.assertEqual(
            team,
            0
        )

        self.assertEqual(
            type,
            0
        )

        self.assertEqual(
            health,
            0
        )

        self.assertIsInstance(
            position,
            Vector3D
        )



class Threshold(object):
    def __init__(self):
        self.health = 50
        self.ammo = 50
        self.aim = 1
        self.shot = 1
        self.look = 1

    def get_health(self):
        """
		Get the stablished limit of health. Agent can perform some actions if
		its health is lower than this value.

		Returns:
			health: current threshold for health
		"""
        return self.health

    def get_ammo(self):
        """
		Get the stablished limit of ammunition. Agent can perform some actions if its ammo is lower than this value.

		Returns:
			ammo: current threshold for ammo

		"""
        return self.ammo

    def get_aim(self):
        """
		Get the stablished number of times that the agent must aim the enemy
		before to shoot.

		Returns:
			aim: current threshold for aim
		"""
        return self.aim

    def get_shot(self):
        """
		Get the stablished number of times that the agent must shoot
		consecutively before doing other action.

		Returns:
			shot: current threshold for shot
		"""
        return self.shot

    def get_look(self):
        """
		Get the stablished number of times (cycles) that the agent must wait
		(moving blindly) before looking again.

		Returns:
			look: current threshold for look

		"""
        return self.look

    def set_health(self, health):
        """
		Stablish the limit of health. Agent can perform some actions if its
		health is lower than this value.
		Rank is [0..100].

		Args:
			health (int):health desired threshold for health
		"""
        if health > 100:
            health = 100
        if health < 0:
            health = 0
        self.health = health

    def set_ammo(self, ammo):
        """
		Stablish the limit of ammunition. Agent can perform some actions if its
		ammo is lower than this value.
		Rank is [0..100].

		Args:
			ammo desired threshold for ammo
		"""
        if ammo > 100:
            ammo = 100
        if ammo < 0:
            ammo = 0
        self.ammo = ammo

    def set_aim(self, aim):
        """
		Stablish the number of times that the agent must aim the enemy before
		to shoot. Rank is [1..20].

		Args:
			aim: desired threshold for aim
		"""
        if aim > 20:
            aim = 20
        if aim < 1:
            aim = 1
        self.aim = aim

    def set_shot(self, shot):
        """
		Stablish the number of times that the agent must shoot consecutively
		before doing other action. Rank is [1..20].

		Args:
			shot:  desired threshold for shot
		"""
        if shot > 20:
            shot = 20
        if shot < 1:
            shot = 1
        self.shot = shot

    def set_look(self, look):
        """
		Stablish the number of times (cycles) that the agent must wait (moving
		blindly) before looking again. Rank is [0..100].

		Args:
			look: desired threshold for look
		"""
        if look > 100:
            look = 100
        if look < 0:
            look = 0
        self.look = look



import asyncio
import struct

import msgpack

from loguru import logger

from spade.container import Container

TCP_COM = 0  # COMMUNICATION (ACCEPTED, CLOSED, REFUSED)
TCP_AGL = 1  # AGENT LIST
TCP_MAP = 2  # MAP: NAME, CHANGES, etc.
TCP_TIME = 3  # TIME: LEFT TIME
TCP_ERR = 4  # ERROR

MSG_TYPE = 501
MSG_BODY = 502
WELCOME_MSG = 503
READY_MSG = 504
QUIT_MSG = 505
ACCEPT_MSG = 506

MSG_AGENTS = 1001
MSG_PACKS = 1002
MSG_CONTENT_NAME = 1003
MSG_CONTENT_TYPE = 1004
MSG_CONTENT_TEAM = 1005
MSG_CONTENT_HEALTH = 1006
MSG_CONTENT_AMMO = 1007
MSG_CONTENT_POSITION = 1008
MSG_CONTENT_VELOCITY = 1009
MSG_CONTENT_HEADING = 1010
MSG_CONTENT_CARRYINGFLAG = 1011


class Server(object):
    def __init__(self, map_name, port=8001):
        self.clients = {}
        self.map_name = map_name
        self.port = port
        self.server = None

        self.container = Container()

        self.loop = self.container.loop

        self.coro = asyncio.start_server(
            self.accept_client, "", self.port, loop=self.loop
        )

    def get_connections(self):
        return self.clients.keys()

    def start(self):
        self.server = self.loop.create_task(self.coro)
        logger.info("Render Server started: {}".format(self.server))

    def stop(self):
        self.server.stop()
        self.loop.run_until_complete(self.server.wait_closed())

    def accept_client(self, client_reader, client_writer):
        logger.info("New render connection")
        task = asyncio.Task(self.handle_client(client_reader, client_writer))
        self.clients[task] = (client_reader, client_writer, False)

        def client_done(task_):
            del self.clients[task_]
            client_writer.close()
            logger.info("End Connection")

        task.add_done_callback(client_done)

    def is_ready(self, task):
        return self.clients[task][2]

    async def handle_client(self, reader, writer):
        task = None
        for k, v in self.clients.items():
            if v[0] == reader and v[1] == writer:
                task = k
                break
        # + ":" + str(self.request)
        logger.info("Preparing Connection to " + str(task))

        try:
            self.send_msg_to_render_engine(task, msg_type=TCP_COM, msg=WELCOME_MSG)
            await writer.drain()
            logger.info("pygomas render engine server v. 0.2.0")
        except Exception as e:
            logger.error("EXCEPTION IN WELCOME MESSAGE")
            logger.error(str(e))

        while True:
            size_of_msg = await reader.read(4)
            if not size_of_msg:
                continue
            size_of_msg = struct.unpack(">I", size_of_msg)[0]
            logger.debug("Got size " + str(size_of_msg))

            data = bytes()
            while len(data) < size_of_msg:
                packet = await reader.read(size_of_msg - len(data))
                if not packet:
                    break
                data += packet

            if len(data) == 0:
                logger.info("Received no data")
                # exit loop and disconnect
                return

            data = msgpack.unpackb(data, raw=False)

            logger.info("Client says:" + str(data))
            if data[MSG_TYPE] == TCP_COM:
                if data[MSG_BODY] == READY_MSG:
                    logger.info("Server: Connection Accepted")
                    self.send_msg_to_render_engine(task, TCP_COM, ACCEPT_MSG)
                    self.send_msg_to_render_engine(task, TCP_MAP, self.map_name)
                    logger.info("Sending: NAME: " + self.map_name)

                    self.clients[task] = (reader, writer, True)

                elif data[MSG_BODY] == QUIT_MSG:
                    logger.info("Server: Client quitted")
                    self.send_msg_to_render_engine(
                        task, TCP_COM, "Server: Connection Closed"
                    )
                    return
                else:
                    # Close connection
                    logger.info("Socket closed, closing connection.")
                    return

            elif data[MSG_TYPE] == TCP_MAP:
                logger.info("Server: Client requested mapname")
                self.send_msg_to_render_engine(task, TCP_MAP, self.map_name)
                self.clients[task] = (reader, writer, True)

    def send_msg_to_render_engine(self, task, msg_type, msg):
        writer = None
        for k, v in self.clients.items():
            if k == task:
                writer = v[1]
                break
        if writer is None:
            logger.info("Connection for {task} not found".format(task=task))
            return

        msg_to_send = msgpack.packb(
            {MSG_TYPE: msg_type, MSG_BODY: msg}, use_bin_type=True
        )
        size_of_package = len(msg_to_send)

        try:
            msg = struct.pack(">I", size_of_package) + msg_to_send
            writer.write(msg)
        except Exception as e:
            logger.error("EXCEPTION IN SENDMSGTORE: {}".format(e))



import os

import numpy as np
from loguru import logger

from .vector import Vector3D

MAP_SCALE = 8


class Base:
    def __init__(self):
        self.init = Vector3D()
        self.end = Vector3D()

    def get_init_x(self):
        return self.init.x

    def get_init_y(self):
        return self.init.y

    def get_init_z(self):
        return self.init.z

    def get_end_x(self):
        return self.end.x

    def get_end_y(self):
        return self.end.y

    def get_end_z(self):
        return self.end.z


class TerrainMap:
    """
    The terrain map is an array of locations. 
    Each location is in turn an array, as follows:
    location[0] = height,
    location[1] = walkable,
    location[2] = cost,
    """

    def __init__(self):
        self.allied_base = Base()
        self.axis_base = Base()
        self.target = Vector3D()

        self.size_x = 0
        self.size_z = 0

    def get_size_x(self):
        return self.size_x

    def get_size_z(self):
        return self.size_z

    def get_target_x(self):
        return self.target.x

    def get_target_y(self):
        return self.target.y

    def get_target_z(self):
        return self.target.z

    def can_walk(self, x, z):
        if (
            x < 0
            or z < 0
            or x >= self.size_x
            or z >= self.size_z
            or len(self.terrain) == 0
        ):
            logger.info(
                "Can't walk outside the map! {}:{} <> {}:{}".format(
                    x, z, self.size_x, self.size_z
                )
            )
            return False

        # logger.info("self.terrain[{}][{}]: {}".format(x, z, self.terrain[x][z].can_walk))
        return self.terrain[x][z][1]

    def get_cost(self, x, z):
        if (
            x < 0
            or z < 0
            or x >= self.size_x
            or z >= self.size_z
            or len(self.terrain) == 0
        ):
            return 2 * 10000

        return self.terrain[x][z][2]

    def load_map(self, main_file, config):

        file = open(config.data_path + main_file + os.sep + main_file + ".txt")

        _ = file.readline()  # [pGomas]
        line = file.readline()  # pGomas_OBJECTIVE:  x y
        tokens = line.split()
        self.target.x = int(tokens[1]) * MAP_SCALE  # x
        self.target.y = 0
        self.target.z = int(tokens[2]) * MAP_SCALE  # y
        logger.info(f" OBJECTIVE: ({self.target.x}, {self.target.z})")

        line = file.readline()  # pGomas_SPAWN_ALLIED: x1 y1 x2 y2
        tokens = line.split()
        self.allied_base.init.x = int(tokens[1]) * MAP_SCALE  # x1
        self.allied_base.init.z = int(tokens[2]) * MAP_SCALE  # z1
        self.allied_base.end.x = int(tokens[3]) * MAP_SCALE  # x2
        self.allied_base.end.z = int(tokens[4]) * MAP_SCALE  # z2
        logger.info(
            f" ALLIED BASE: ({self.allied_base.init.x}, {self.allied_base.init.z})"
            f" ({self.allied_base.end.x}, {self.allied_base.end.z})"
        )

        line = file.readline()  # pGomas_SPAWN_AXIS: x1 y1 x2 y2
        tokens = line.split()
        self.axis_base.init.x = int(tokens[1]) * MAP_SCALE  # x1
        self.axis_base.init.z = int(tokens[2]) * MAP_SCALE  # z1
        self.axis_base.end.x = int(tokens[3]) * MAP_SCALE  # x2
        self.axis_base.end.z = int(tokens[4]) * MAP_SCALE  # z2
        logger.info(
            f" AXIS BASE: ({self.axis_base.init.x}, {self.axis_base.init.z}) "
            f"({self.axis_base.end.x}, {self.axis_base.end.z})"
        )

        line = file.readline()  # pGomas_COST_MAP: w h name
        tokens = line.split()
        self.size_x = int(tokens[1]) * MAP_SCALE
        self.size_z = int(tokens[2]) * MAP_SCALE
        cost_map_name = tokens[3]
        logger.info(f" COST MAP: ({self.size_x} x {self.size_z}) {cost_map_name}")

        file.close()

        if self.size_x <= 0 or self.size_z <= 0:
            logger.info("Invalid Cost Map")
            return

        self.terrain = np.zeros(shape=(self.size_x, self.size_z, 3))

        file = open(config.data_path + main_file + os.sep + cost_map_name)

        for z in range(self.size_z // MAP_SCALE):
            for x in range(self.size_x // MAP_SCALE):
                c = file.read(1)
                while c == "\n" or c == "\r":
                    c = file.read(1)  # read next char
                if c == "*":
                    self.terrain[
                        x * MAP_SCALE : x * MAP_SCALE + MAP_SCALE,
                        z * MAP_SCALE : z * MAP_SCALE + MAP_SCALE,
                        0,
                    ] = 0.0
                    self.terrain[
                        x * MAP_SCALE : x * MAP_SCALE + MAP_SCALE,
                        z * MAP_SCALE : z * MAP_SCALE + MAP_SCALE,
                        1,
                    ] = 0
                    self.terrain[
                        x * MAP_SCALE : x * MAP_SCALE + MAP_SCALE,
                        z * MAP_SCALE : z * MAP_SCALE + MAP_SCALE,
                        2,
                    ] = 10000
                elif c == " ":
                    self.terrain[
                        x * MAP_SCALE : x * MAP_SCALE + MAP_SCALE,
                        z * MAP_SCALE : z * MAP_SCALE + MAP_SCALE,
                        0,
                    ] = 0.0
                    self.terrain[
                        x * MAP_SCALE : x * MAP_SCALE + MAP_SCALE,
                        z * MAP_SCALE : z * MAP_SCALE + MAP_SCALE,
                        1,
                    ] = 1
                    self.terrain[
                        x * MAP_SCALE : x * MAP_SCALE + MAP_SCALE,
                        z * MAP_SCALE : z * MAP_SCALE + MAP_SCALE,
                        2,
                    ] = 1
        file.close()

        self.cost_terrain = np.copy(self.terrain)

        for z in range(1, self.size_z - 1):
            for x in range(1, self.size_x - 1):
                if self.terrain[x, z, 1] == 1:
                    if (
                        self.terrain[x - 1, z - 1, 1] == 0
                        or self.terrain[x, z - 1, 1] == 0
                        or self.terrain[x + 1, z - 1, 1] == 0
                        or self.terrain[x - 1, z, 1] == 0
                        or self.terrain[x + 1, z, 1] == 0
                        or self.terrain[x - 1, z + 1, 1] == 0
                        or self.terrain[x, z + 1, 1] == 0
                        or self.terrain[x + 1, z + 1, 1] == 0
                    ):
                        self.cost_terrain[x, z, 1] = 0
                        self.cost_terrain[x, z, 2] = 5000

    def __str__(self):
        s = ""

        for z in range(self.size_z):
            for x in range(self.size_x):
                s += str(self.terrain[x][z])
            s += "\n"
        return s



import unittest
from jgomas.CObjPack import CObjPack
from jgomas.CJGomasAgent import CJGomasAgent
from spade.behaviour import OneShotBehaviour
from spade.message import Message
from spade.template import Template
import time


class TestCObjPack(unittest.TestCase):

    def test_setup(self):
        class TestBehaviour(OneShotBehaviour):
            async def run(self):
                res = await self.receive(timeout=10000)
                msg = Message(to="objective1@localhost")
                msg.set_metadata("performative", "inform")
                msg.body = "Hello"
                await self.send(msg)

        agent = CJGomasAgent('cmanager@localhost')
        agent.start()

        behav = TestBehaviour()
        agent.add_behaviour(behav)

        pack = CObjPack("objective1@localhost")

        pack.start()

        while not behav.done():
            time.sleep(1)

        agent.stop()

        pack.stop()



import math


class Vector3D(object):

    # Vector length
    def length(self):
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def __len__(self):
        return self.length()

    # Vector addition
    def add(self, v):
        self.x = self.x + v.x
        self.y = self.y + v.y
        self.z = self.z + v.z

    # Vector subtraction
    def sub(self, v):
        self.x = self.x - v.x
        self.y = self.y - v.y
        self.z = self.z - v.z

    # Dot product
    def dot(self, v):
        return self.x * v.x + self.y * v.y + self.z * v.z

    # Vector normalization
    def normalize(self):
        l = self.length()
        if l == 0:
            return
        self.x = self.x / l
        self.y = self.y / l
        self.z = self.z / l

    # Cross product
    def cross(self, v):
        u = Vector3D()
        u.x = self.y * v.z - self.z * v.y
        u.y = self.z * v.x - self.x * v.z
        u.z = self.x * v.y - self.y * v.x
        return u

    def __eq__(self, other):
        if self.x == other.x and self.y == other.y and self.z == other.z:
            return True
        else:
            return False

    def __ne__(self, other):
        if self.x == other.x and self.y == other.y and self.z == other.z:
            return False
        else:
            return True

    def __str__(self):
        return "<" + str(self.x) + "," + str(self.y) + "," + str(self.z) + ">"

    def __init__(self, v=None, x=None, y=None, z=None):
        if v is None and not x and not y and not z:
            # New vector
            self.x = self.y = self.z = 0.0
        elif v is not None:
            # From another vector
            self.x = v.x
            self.y = v.y
            self.z = v.z
        elif x or y or z:
            # By coordinates
            if x:
                self.x = x
            else:
                self.x = 0.0
            if y:
                self.y = y
            else:
                self.y = 0.0
            if z:
                self.z = z
            else:
                self.z = 0.0



import json
from loguru import logger

from .ontology import (
    ACTION,
    CREATE,
    TYPE,
    X,
    Y,
    Z,
    PERFORMATIVE,
    PERFORMATIVE_PACK,
    NAME,
    PERFORMATIVE_PACK_TAKEN,
    TEAM,
)
from .agent import AbstractAgent, LONG_RECEIVE_WAIT
from .vector import Vector3D
from spade.message import Message
from spade.behaviour import OneShotBehaviour, CyclicBehaviour
from spade.template import Template
from spade.agent import Agent

PACK_NONE: int = 1000
PACK_MEDICPACK: int = 1001
PACK_AMMOPACK: int = 1002
PACK_OBJPACK: int = 1003

PACK_NAME = {
    PACK_NONE: "NONE",
    PACK_MEDICPACK: "MEDIC",
    PACK_AMMOPACK: "AMMO",
    PACK_OBJPACK: "OBJ",
}

PACK_AUTODESTROY_TIMEOUT: int = 25


class Pack(AbstractAgent, Agent):
    def __str__(self):
        return "P(" + str(PACK_NAME[self.type]) + "," + str(self.position) + ")"

    def __init__(
        self, name, passwd="secret", manager_jid="cmanager@localhost", x=0, z=0, team=0
    ):
        Agent.__init__(self, name, passwd)
        AbstractAgent.__init__(self, name, team)

        self.type = PACK_NONE
        self.manager = manager_jid

        self.position = Vector3D()
        self.position.x = x
        self.position.y = 0
        self.position.z = z

    async def setup(self):
        self.add_behaviour(self.CreatePackBehaviour())

        t = Template()
        t.set_metadata(PERFORMATIVE, PERFORMATIVE_PACK_TAKEN)
        self.add_behaviour(self.PackTakenResponderBehaviour(), t)

    class CreatePackBehaviour(OneShotBehaviour):
        async def run(self):
            msg = Message(to=self.agent.manager)
            msg.set_metadata(PERFORMATIVE, PERFORMATIVE_PACK)
            msg.body = json.dumps(
                {
                    NAME: self.agent.name,
                    TEAM: self.agent.team,
                    ACTION: CREATE,
                    TYPE: self.agent.type,
                    X: self.agent.position.x,
                    Y: self.agent.position.y,
                    Z: self.agent.position.z,
                }
            )
            await self.send(msg)
            logger.info("CreatePack msg sent: {}".format(msg))

    class PackTakenResponderBehaviour(CyclicBehaviour):
        async def run(self):
            msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
            if msg is not None:
                content = msg.body
                await self.agent.perform_pack_taken(content)
                # await self.agent.stop()



import json

from loguru import logger

from spade.agent import Agent
from spade.behaviour import CyclicBehaviour
from spade.template import Template

from .agent import LONG_RECEIVE_WAIT
from .ontology import (
    PERFORMATIVE,
    PERFORMATIVE_CFA,
    PERFORMATIVE_CFB,
    PERFORMATIVE_CFM,
    PERFORMATIVE_DEREGISTER_AGENT,
    PERFORMATIVE_DEREGISTER_SERVICE,
    PERFORMATIVE_GET,
    PERFORMATIVE_REGISTER_SERVICE,
    AMMO_SERVICE,
    BACKUP_SERVICE,
    MEDIC_SERVICE,
    NAME,
    TEAM,
)
from .config import TEAM_NONE, TEAM_ALLIED, TEAM_AXIS


class Service(Agent):
    def __init__(self, jid="cservice@localhost", password="secret"):
        self.services = {}
        super().__init__(jid=jid, password=password)

    def register_service(self, service_descriptor, jid):
        name = service_descriptor[NAME]
        team = service_descriptor[TEAM]

        if name not in self.services.keys():
            self.services[name] = {TEAM_AXIS: [], TEAM_ALLIED: [], TEAM_NONE: []}

        self.services[name][team].append(jid)
        logger.info("Service {} of team {} registered for {}".format(name, team, jid))

    def deregister_service(self, service_descriptor, jid):
        name = service_descriptor[NAME]
        team = service_descriptor[TEAM]

        if name in self.services.keys() and jid in self.services[name][team]:
            self.services[name][team].remove(jid)
        logger.info("Service {} of team {} deregistered for {}".format(name, team, jid))

    def deregister_agent(self, jid):
        logger.info("Deregistering all services of agent {}".format(jid))
        for name in self.services.keys():
            for team in [TEAM_ALLIED, TEAM_AXIS]:
                if jid in self.services[name][team]:
                    self.services[name][team].remove(jid)
                    logger.info(
                        "Service {} of team {} deregistered for {}".format(
                            name, team, jid
                        )
                    )

    def get_service(self, service_descriptor, questioner):
        logger.debug("get service: {}".format(service_descriptor))
        name = service_descriptor[NAME]
        team = service_descriptor[TEAM]

        if name in self.services.keys():
            logger.debug("I got service")
            request = self.services[name][team][:]
            if questioner in request:
                request.remove(questioner)
            return request
        else:
            logger.debug("No service")
            return []

    async def setup(self):
        template1 = Template()
        template1.set_metadata(PERFORMATIVE, PERFORMATIVE_REGISTER_SERVICE)
        self.add_behaviour(RegisterServiceBehaviour(), template1)

        template2 = Template()
        template2.set_metadata(PERFORMATIVE, PERFORMATIVE_DEREGISTER_SERVICE)
        self.add_behaviour(DeregisterServiceBehaviour(), template2)

        template3 = Template()
        template3.set_metadata(PERFORMATIVE, PERFORMATIVE_DEREGISTER_AGENT)
        self.add_behaviour(DeregisterAgentBehaviour(), template3)

        template4 = Template()
        template4.set_metadata(PERFORMATIVE, PERFORMATIVE_GET)
        self.add_behaviour(GetServiceBehaviour(), template4)


class RegisterServiceBehaviour(CyclicBehaviour):
    async def run(self):
        msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
        if msg:
            logger.info(
                "Register Service {} for {}.".format(msg.body, msg.sender.bare())
            )
            self.agent.register_service(json.loads(msg.body), str(msg.sender.bare()))


class DeregisterServiceBehaviour(CyclicBehaviour):
    async def run(self):
        msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
        if msg:
            logger.info(
                "Deregister Service {} for {}.".format(msg.body, msg.sender.bare())
            )
            self.agent.deregister_service(json.loads(msg.body), str(msg.sender.bare()))


class DeregisterAgentBehaviour(CyclicBehaviour):
    async def run(self):
        msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
        if msg:
            self.agent.deregister_agent(str(msg.sender))
            logger.info("Agent {} deregistered".format(msg.sender))


class GetServiceBehaviour(CyclicBehaviour):
    async def run(self):
        msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
        if msg:
            logger.info("Requesting service {}".format(msg.body))
            body = json.loads(msg.body)
            names = self.agent.get_service(body, str(msg.sender).split("/")[0])
            reply = msg.make_reply()
            reply.body = json.dumps(names)
            if body[NAME] == AMMO_SERVICE:
                reply.set_metadata(PERFORMATIVE, PERFORMATIVE_CFA)
            elif body[NAME] == MEDIC_SERVICE:
                reply.set_metadata(PERFORMATIVE, PERFORMATIVE_CFM)
            elif body[NAME] == BACKUP_SERVICE:
                reply.set_metadata(PERFORMATIVE, PERFORMATIVE_CFB)
            else:
                reply.set_metadata(PERFORMATIVE, body[NAME])
            await self.send(reply)
            logger.info("Services sent: {}".format(reply.body))



import datetime

from .pack import PACK_MEDICPACK, PACK_AMMOPACK, PACK_OBJPACK
from .config import TEAM_ALLIED, TEAM_AXIS


class PackStatistic:
    def __init__(self):
        self.delivered = 0
        self.team_taken = 0
        self.enemy_taken = 0
        self.not_taken = 0

    def __str__(self):
        ret = "\t\t* Delivered:   \t" + str(self.delivered) + "\n"
        ret += "\t\t* Team Taken:  \t" + str(self.team_taken) + "\n"
        ret += "\t\t* Enemy Taken: \t" + str(self.enemy_taken) + "\n"
        ret += "\t\t* Not Taken:   \t" + str(self.not_taken) + "\n"
        return ret


class TeamStatistic:
    def __init__(self):

        self.packs = {PACK_MEDICPACK: PackStatistic(), PACK_AMMOPACK: PackStatistic()}

        self.total_shots = 0
        self.enemy_hit_shots = 0
        self.team_hit_shots = 0
        self.failed_shots = 0

        self.total_objective_taken = 0
        self.total_objective_lost = 0

        self.medic_efficiency = 0
        self.fieldops_efficiency = 0
        self.army_efficiency = 0

        self.medic_anti_efficiency = 0
        self.fieldops_anti_efficiency = 0
        self.army_anti_efficiency = 0

        self.alive_players = 0
        self.average_health = 0.0

    def calculate_efficiency(self, alive_enemies):

        if self.packs[PACK_MEDICPACK].delivered <= 0:
            self.medic_efficiency = 0
        else:
            not_taken = self.packs[PACK_MEDICPACK].not_taken
            delivered = self.packs[PACK_MEDICPACK].delivered
            self.medic_efficiency = 1.0 - (not_taken * 1.0) / delivered

        if self.packs[PACK_AMMOPACK].delivered <= 0:
            self.fieldops_efficiency = 0
        else:
            not_taken = self.packs[PACK_AMMOPACK].not_taken
            delivered = self.packs[PACK_AMMOPACK].delivered
            self.fieldops_efficiency = 1.0 - (not_taken * 1.0) / delivered

        if self.total_shots <= 0:
            self.army_efficiency = 0
        else:
            self.army_efficiency = 1.0 - (alive_enemies * 1.0) / self.total_shots

    def calculate_anti_efficiency(self):

        if self.packs[PACK_MEDICPACK].delivered <= 0:
            self.medic_anti_efficiency = 0
        else:
            enemy_taken = self.packs[PACK_MEDICPACK].enemy_taken
            delivered = self.packs[PACK_MEDICPACK].delivered
            self.medic_anti_efficiency = (enemy_taken * 1.0) / delivered

        if self.packs[PACK_AMMOPACK].delivered <= 0:
            self.fieldops_anti_efficiency = 0
        else:
            enemy_taken = self.packs[PACK_AMMOPACK].enemy_taken
            delivered = self.packs[PACK_AMMOPACK].delivered
            self.fieldops_anti_efficiency = (enemy_taken * 1.0) / delivered

        if self.team_hit_shots <= 0:
            self.army_anti_efficiency = 0
        else:
            self.army_anti_efficiency = (self.alive_players * 1.0) / self.team_hit_shots

    def __str__(self):

        ret = "\t-GENERAL:\n"
        ret += "\t\t* Alive:       \t" + str(self.alive_players) + "\n"
        ret += "\t\t* Avrg. Health:\t" + str(self.average_health) + "\n"

        ret += "\t-OBJECTIVE:\n"
        ret += "\t\t* Times Taken: \t" + str(self.total_objective_taken) + "\n"
        ret += "\t\t* Times Lost:  \t" + str(self.total_objective_lost) + "\n"

        ret += "\t-SHOTS:\n"
        ret += "\t\t* EnemyHit:    \t" + str(self.enemy_hit_shots) + "\n"
        ret += "\t\t* TeamHit:     \t" + str(self.team_hit_shots) + "\n"
        ret += "\t\t* FailedHit:   \t" + str(self.failed_shots) + "\n"
        ret += "\t\t* TOTAL:       \t" + str(self.total_shots) + "\n"

        ret += "\t-MEDIC PACKS:\n"
        ret += str(self.packs[PACK_MEDICPACK]) + "\n"

        ret += "\t-AMMO PACKS:\n"
        ret += str(self.packs[PACK_AMMOPACK]) + "\n"

        ret += "\t-EFICIENCY:\n"
        ret += "\t\t* Medic:       \t" + str(self.medic_efficiency) + "\n"
        ret += "\t\t* FieldOps:    \t" + str(self.fieldops_efficiency) + "\n"
        ret += "\t\t* Army:        \t" + str(self.army_efficiency) + "\n"

        ret += "\t-ANTI-EFICIENCY:" + "\n"
        ret += "\t\t* Medic:       \t" + str(self.medic_anti_efficiency) + "\n"
        ret += "\t\t* FieldOps:    \t" + str(self.fieldops_anti_efficiency) + "\n"
        ret += "\t\t* Army:        \t" + str(self.army_anti_efficiency) + "\n"

        return ret


class GameStatistic:
    def __init__(self):
        self.team_statistic = {TEAM_ALLIED: TeamStatistic(), TEAM_AXIS: TeamStatistic()}

        self.match_duration = 0

    def pack_taken(self, pack, team):
        pack_team = TEAM_ALLIED if pack.team == TEAM_ALLIED else TEAM_AXIS

        if pack.type == PACK_OBJPACK:
            if team == TEAM_ALLIED:
                self.team_statistic[TEAM_ALLIED].total_objective_taken += 1
                self.team_statistic[TEAM_AXIS].total_objective_lost += 1
            elif team == TEAM_AXIS:
                self.team_statistic[TEAM_AXIS].total_objective_taken += 1
        else:
            if pack.team == team:
                self.team_statistic[pack_team].packs[pack.type].team_taken += 1
            else:
                self.team_statistic[pack_team].packs[pack.type].enemy_taken += 1

    def shoot(self, victim, team):
        team = TEAM_ALLIED if team == TEAM_ALLIED else TEAM_AXIS
        self.team_statistic[team].total_shots += 1

        if victim is None:
            self.team_statistic[team].failed_shots += 1
        elif team == victim.team:
            self.team_statistic[team].team_hit_shots += 1
        else:
            self.team_statistic[team].enemy_hit_shots += 1

    def objective_lost(self, team):
        team = TEAM_ALLIED if team == TEAM_ALLIED else TEAM_AXIS
        self.team_statistic[team].total_objective_lost += 1

    def pack_destroyed(self, din_object):
        if din_object.team == TEAM_ALLIED:
            pack_team = TEAM_ALLIED
        else:
            pack_team = TEAM_AXIS
        pack_type = -1
        if din_object.type == PACK_MEDICPACK:
            pack_type = PACK_MEDICPACK
        elif din_object.type == PACK_AMMOPACK:
            pack_type = PACK_AMMOPACK
        if pack_type >= 0:
            self.team_statistic[pack_team].packs[pack_type].not_taken += 1

    def pack_created(self, din_object, team):
        pack_team = TEAM_ALLIED if team == TEAM_ALLIED else TEAM_AXIS
        pack_type = -1
        if din_object.type == PACK_MEDICPACK:
            pack_type = PACK_MEDICPACK
        elif din_object.type == PACK_AMMOPACK:
            pack_type = PACK_AMMOPACK

        if pack_type >= 0:
            self.team_statistic[pack_team].packs[pack_type].delivered += 1

    def calculate_data(
        self, allied_alive_players, axis_alive_players, allied_health, axis_health
    ):

        self.team_statistic[TEAM_ALLIED].alive_players = allied_alive_players
        if allied_alive_players > 0:
            self.team_statistic[TEAM_ALLIED].average_health = (
                allied_health * 1.0
            ) / allied_alive_players
        else:
            self.team_statistic[TEAM_ALLIED].average_health = 0

        self.team_statistic[TEAM_AXIS].alive_players = axis_alive_players
        if axis_alive_players > 0:
            self.team_statistic[TEAM_AXIS].average_health = (
                axis_health * 1.0
            ) / axis_alive_players
        else:
            self.team_statistic[TEAM_AXIS].average_health = 0

        self.team_statistic[TEAM_ALLIED].calculate_efficiency(axis_alive_players)
        self.team_statistic[TEAM_ALLIED].calculate_anti_efficiency()

        self.team_statistic[TEAM_AXIS].calculate_efficiency(allied_alive_players)
        self.team_statistic[TEAM_AXIS].calculate_anti_efficiency()

    def dumps(self, winner_team=""):

        ret = "Winner Team: " + winner_team + "\n"
        ret += (
            "Duration: ["
            + str(str(datetime.timedelta(seconds=self.match_duration)))
            + "]\n"
        )
        ret += "Statistics for ALLIED TEAM\n"
        ret += str(self.team_statistic[TEAM_ALLIED])

        ret += "\n"
        ret += "Statistics for AXIS TEAM\n"
        ret += str(self.team_statistic[TEAM_AXIS])

        ret = ret + "\n"
        return ret



import json
from abc import ABCMeta

from loguru import logger

from spade.agent import Agent
from spade.behaviour import OneShotBehaviour
from spade.message import Message

from .ontology import (
    PERFORMATIVE,
    PERFORMATIVE_DEREGISTER_AGENT,
    PERFORMATIVE_DEREGISTER_SERVICE,
    PERFORMATIVE_REGISTER_SERVICE,
    NAME,
    TEAM,
)

LONG_RECEIVE_WAIT: int = 1000000


class AbstractAgent(object, metaclass=ABCMeta):
    def __init__(self, jid, team=0, service_jid="cservice@localhost"):
        self.services = list()
        self._name = jid
        self.team = team
        self.service_jid = service_jid
        self.alive = True

    def start(self, auto_register=True):
        future = Agent.start(self, auto_register=auto_register)
        if self.services:
            for service in self.services:
                logger.info("{} registering service {}".format(self.name, service))
                self.register_service(service)
        return future

    async def die(self):
        await self.deregister_agent()
        self.alive = False
        await self.stop()
        logger.info("Agent {} was stopped.".format(self.name))

    async def send(self, msg):
        if self.is_alive():
            await super().send(msg)

    def register_service(self, service_name):
        class RegisterBehaviour(OneShotBehaviour):
            async def run(self):
                msg = Message(to=self.agent.service_jid)
                msg.set_metadata(PERFORMATIVE, PERFORMATIVE_REGISTER_SERVICE)
                msg.body = json.dumps({NAME: service_name, TEAM: self.agent.team})
                await self.send(msg)

        self.add_behaviour(RegisterBehaviour())

    def deregister_service(self, service_name):
        class DeregisterBehaviour(OneShotBehaviour):
            async def run(self):
                msg = Message(to=self.agent.service_jid)
                msg.set_metadata(PERFORMATIVE, PERFORMATIVE_DEREGISTER_SERVICE)
                msg.body = json.dumps({NAME: service_name, TEAM: self.agent.team})
                await self.send(msg)

        self.add_behaviour(DeregisterBehaviour())

    async def deregister_agent(self):
        class DeregisterAgentBehaviour(OneShotBehaviour):
            async def run(self):
                msg = Message(to=self.agent.service_jid)
                msg.set_metadata(PERFORMATIVE, PERFORMATIVE_DEREGISTER_AGENT)
                await self.send(msg)

        behav = DeregisterAgentBehaviour()
        self.add_behaviour(behav)
        # await behav.join(timeout=5)

    @property
    def name(self):
        return self._name



import asyncio
import json
from spade.behaviour import TimeoutBehaviour

from .ontology import ACTION, DESTROY, PERFORMATIVE, PERFORMATIVE_INFORM, NAME
from .pack import Pack, PACK_AMMOPACK, PACK_AUTODESTROY_TIMEOUT
from datetime import datetime, timedelta
from spade.message import Message

now = datetime.now


class AmmoPack(Pack):
    async def setup(self):
        self.type = PACK_AMMOPACK
        timeout = now() + timedelta(seconds=PACK_AUTODESTROY_TIMEOUT)
        self.add_behaviour(self.AutoDestroyBehaviour(start_at=timeout))
        await super().setup()

    async def perform_pack_taken(self, content):
        await self.stop()

    class AutoDestroyBehaviour(TimeoutBehaviour):
        async def run(self):
            msg = Message(to=self.agent.manager)
            msg.set_metadata(PERFORMATIVE, PERFORMATIVE_INFORM)
            content = {NAME: self.agent.name, ACTION: DESTROY}
            msg.body = json.dumps(content)
            if self.agent.is_alive():
                await self.send(msg)
            await asyncio.sleep(1)
            await self.agent.stop()



import unittest
from pygomas.CConfig import CConfig


class TestCConfig(unittest.TestCase):

    def test_set_data_path(self):
        config = CConfig()

        self.assertEqual(
            config.m_dataPath,
            "../../../../data/maps/"
        )

        config.setDataPath("./data")

        self.assertEqual(
            config.m_dataPath,
            "./data/"
        )



import asyncio
import datetime
import json
import math
import random
import time
import traceback

import spade
from loguru import logger
from spade.agent import Agent
from spade.behaviour import (
    OneShotBehaviour,
    PeriodicBehaviour,
    CyclicBehaviour,
    TimeoutBehaviour,
)
from spade.message import Message
from spade.template import Template

from . import __version__
from .agent import AbstractAgent, LONG_RECEIVE_WAIT
from .bditroop import CLASS_SOLDIER
from .config import (
    Config,
    MIN_HEALTH,
    TEAM_NONE,
    TEAM_ALLIED,
    TEAM_AXIS,
    MISSING_SHOT_PROBABILITY,
)
from .map import TerrainMap
from .mobile import Mobile
from .objpack import ObjectivePack
from .ontology import (
    ACTION,
    AIM,
    ANGLE,
    CREATE,
    DEC_HEALTH,
    DESTROY,
    DISTANCE,
    FOV,
    HEAD_X,
    HEAD_Y,
    HEAD_Z,
    MAP,
    PACKS,
    QTY,
    SHOTS,
    VEL_X,
    TYPE,
    VEL_Y,
    VEL_Z,
    X,
    Y,
    Z,
    PERFORMATIVE,
    PERFORMATIVE_DATA,
    PERFORMATIVE_GAME,
    PERFORMATIVE_INIT,
    PERFORMATIVE_OBJECTIVE,
    PERFORMATIVE_PACK,
    PERFORMATIVE_PACK_LOST,
    PERFORMATIVE_SHOOT,
    MANAGEMENT_SERVICE,
    AMMO,
    HEALTH,
    NAME,
    PERFORMATIVE_PACK_TAKEN,
    TEAM,
)
from .pack import PACK_NAME, PACK_NONE, PACK_OBJPACK, PACK_MEDICPACK, PACK_AMMOPACK
from .server import (
    Server,
    TCP_AGL,
    TCP_COM,
    MSG_AGENTS,
    MSG_CONTENT_NAME,
    MSG_CONTENT_TYPE,
    MSG_CONTENT_TEAM,
    MSG_CONTENT_HEALTH,
    MSG_CONTENT_AMMO,
    MSG_CONTENT_CARRYINGFLAG,
    MSG_CONTENT_POSITION,
    MSG_CONTENT_VELOCITY,
    MSG_CONTENT_HEADING,
    MSG_PACKS,
    QUIT_MSG,
)
from .service import Service
from .sight import Sight
from .stats import GameStatistic
from .vector import Vector3D

MILLISECONDS_IN_A_SECOND: int = 1000

DEFAULT_PACK_QTY: int = 20

ARG_PLAYERS: int = 0
ARG_MAP_NAME: int = 1
ARG_FPS: int = 2
ARG_MATCH_TIME: int = 3
ARG_MAP_PATH: int = 4

WIDTH: int = 3


class MicroAgent:
    def __init__(self):
        self.jid = ""
        self.team = TEAM_NONE
        self.locate = Mobile()
        self.is_carrying_objective = False
        self.is_shooting = False
        self.health = 0
        self.ammo = 0
        self.type = 0
        self.is_updated = False

    def __str__(self):
        return "<{} Team({}) Health({}) Ammo({}) Obj({})>".format(
            self.jid, self.team, self.health, self.ammo, self.is_carrying_objective
        )


class DinObject:
    def __str__(self):
        return "DO({},{})".format(PACK_NAME[self.type], self.position)

    def __init__(self):
        self.position = Vector3D()
        self.type = PACK_NONE
        self.team = TEAM_NONE
        self.is_taken = False
        self.owner = 0
        self.jid = None


class Manager(AbstractAgent, Agent):
    def __init__(
        self,
        name="cmanager@localhost",
        passwd="secret",
        players=10,
        fps=33,
        match_time=380,
        map_name="map_01",
        map_path=None,
        service_jid="cservice@localhost",
        service_passwd="secret",
        port=8001,
    ):

        AbstractAgent.__init__(self, name, service_jid=service_jid)
        Agent.__init__(self, name, passwd)

        self.game_statistic = GameStatistic()
        self.max_total_agents = players
        self.fps = 1 / fps
        self.match_time = match_time
        self.map_name = str(map_name)
        self.port = port
        self.config = Config(data_path=map_path)
        self.number_of_agents = 0
        self.agents = {}
        self.match_init = 0
        self.domain = name.split("@")[1]
        self.objective_agent = None
        self.service_agent = Service(jid=self.service_jid, password=service_passwd)
        self.render_server = Server(map_name=self.map_name, port=self.port)
        self.din_objects = dict()
        self.map = TerrainMap()

    def _in_coroutine(self):
        try:
            return asyncio.get_event_loop() == self.loop
        except RuntimeError:  # pragma: no cover
            return False

    def stop(self):
        del self.render_server
        self.render_server = None
        if self._in_coroutine():
            coro1 = self.objective_agent.stop()
            coro2 = self.service_agent.stop()
            coro3 = super().stop()
            coro = asyncio.gather(coro1, coro2, coro3)

        else:
            self.objective_agent.stop().result()
            self.service_agent.stop().result()
            coro = super().stop()

        return coro

    async def setup(self):
        class InitBehaviour(OneShotBehaviour):
            async def run(self):
                logger.success(
                    "Manager (Expected Agents): {}".format(self.agent.max_total_agents)
                )

                # for i in range(1, self.agent.max_total_agents + 1):
                while self.agent.number_of_agents < self.agent.max_total_agents:
                    msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
                    if msg:
                        content = json.loads(msg.body)

                        name = content[NAME]
                        type_ = content[TYPE]
                        team = content[TEAM]

                        self.agent.agents[name] = MicroAgent()

                        self.agent.agents[name].jid = name
                        self.agent.agents[name].type = int(type_)
                        self.agent.agents[name].team = int(team)
                        self.agent.agents[name].health = 100

                        logger.success("Manager: [" + name + "] is Ready!")
                        self.agent.number_of_agents += 1

                logger.success(
                    "Manager (Accepted Agents): " + str(self.agent.number_of_agents)
                )
                for agent in self.agent.agents.values():
                    msg = Message()
                    msg.set_metadata(PERFORMATIVE, PERFORMATIVE_INIT)
                    msg.to = agent.jid
                    msg.body = json.dumps({MAP: self.agent.map_name})
                    await self.send(msg)
                    logger.success(
                        "Manager: Sending notification to fight to: " + agent.jid
                    )

                await self.agent.inform_objectives(self)
                self.agent.match_init = time.time()

                # Behaviour to check if all alied troops are alive
                self.agent.launch_check_allied_health()

                # Behaviour to refresh all render engines connected
                self.agent.launch_render_engine_inform_behaviour()

        logger.success(
            "pygomas {} (c) VRAIN 2005-{} (VRAIN/UPV)".format(
                __version__, time.strftime("%Y")
            )
        )
        logger.success("Powered by SPADE {}".format(spade.__version__))

        coro = self.service_agent.start(auto_register=True)
        await coro

        self.register_service(MANAGEMENT_SERVICE)

        self.render_server.start()

        self.map.load_map(self.map_name, self.config)

        # Behaviour to listen to data (position, health?, an so on) from troop agents
        self.launch_data_from_troop_listener_behaviour()

        # Behaviour to handle Shot messages
        self.launch_shoot_responder_behaviour()

        # Behaviour to handle Pack Management: Creation and Destruction
        self.launch_pack_management_responder_behaviour()

        # Behaviour to inform all agents that game has finished by time
        self.launch_game_timeout_inform_behaviour()

        template = Template()
        template.set_metadata(PERFORMATIVE, PERFORMATIVE_INIT)
        self.add_behaviour(InitBehaviour(), template)

        await self.create_objectives()  # We need to do this when online

    # Behaviour to refresh all render engines connected
    def launch_render_engine_inform_behaviour(self):
        class InformRenderEngineBehaviour(PeriodicBehaviour):
            async def run(self):
                if (
                    self.agent.render_server
                    and len(self.agent.render_server.get_connections()) != 0
                    and all([[a.is_updated for a in self.agent.agents.values()]])
                    and len(self.agent.agents) == self.agent.number_of_agents
                ):

                    msg = {
                        MSG_AGENTS: [
                            {
                                MSG_CONTENT_NAME: agent.jid.split("@")[0],
                                MSG_CONTENT_TYPE: agent.type,
                                MSG_CONTENT_TEAM: agent.team,
                                MSG_CONTENT_HEALTH: agent.health,
                                MSG_CONTENT_AMMO: agent.ammo,
                                MSG_CONTENT_CARRYINGFLAG: agent.is_carrying_objective,
                                MSG_CONTENT_POSITION: [
                                    agent.locate.position.x,
                                    agent.locate.position.y,
                                    agent.locate.position.z,
                                ],
                                MSG_CONTENT_VELOCITY: [
                                    agent.locate.velocity.x,
                                    agent.locate.velocity.y,
                                    agent.locate.velocity.z,
                                ],
                                MSG_CONTENT_HEADING: [
                                    agent.locate.heading.x,
                                    agent.locate.heading.y,
                                    agent.locate.heading.z,
                                ],
                            }
                            for agent in self.agent.agents.values()  # if agent.is_updated
                        ],
                        MSG_PACKS: [
                            {
                                MSG_CONTENT_NAME: pack.render_id,
                                MSG_CONTENT_TYPE: pack.type,
                                MSG_CONTENT_POSITION: [
                                    pack.position.x,
                                    pack.position.y,
                                    pack.position.z,
                                ],
                            }
                            for pack in self.agent.din_objects.values()
                            if not pack.is_taken
                        ],
                    }

                    for task in self.agent.render_server.get_connections():
                        if self.agent.render_server.is_ready(task):
                            self.agent.render_server.send_msg_to_render_engine(
                                task, TCP_AGL, msg
                            )

        self.add_behaviour(InformRenderEngineBehaviour(self.fps))
        logger.debug("InformRenderEngineBehaviour started.")

    # Behaviour to listen to data (position, health?, an so on) from troop agents
    def launch_data_from_troop_listener_behaviour(self):
        class DataFromTroopBehaviour(CyclicBehaviour):
            async def run(self):
                msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
                if self.mailbox_size() > self.agent.max_total_agents + 1:
                    logger.error("TOO MUCH PENDING MSG: {}".format(self.mailbox_size()))
                try:
                    if msg:
                        content = json.loads(msg.body)
                        id_agent = content[NAME]

                        self.agent.agents[id_agent].locate.position.x = int(content[X])
                        self.agent.agents[id_agent].locate.position.y = int(content[Y])
                        self.agent.agents[id_agent].locate.position.z = int(content[Z])
                        self.agent.agents[id_agent].is_updated = True

                        self.agent.agents[id_agent].locate.velocity.x = float(
                            content[VEL_X]
                        )
                        self.agent.agents[id_agent].locate.velocity.y = float(
                            content[VEL_Y]
                        )
                        self.agent.agents[id_agent].locate.velocity.z = float(
                            content[VEL_Z]
                        )

                        self.agent.agents[id_agent].locate.heading.x = float(
                            content[HEAD_X]
                        )
                        self.agent.agents[id_agent].locate.heading.y = float(
                            content[HEAD_Y]
                        )
                        self.agent.agents[id_agent].locate.heading.z = float(
                            content[HEAD_Z]
                        )

                        self.agent.agents[id_agent].health = int(content[HEALTH])
                        self.agent.agents[id_agent].ammo = int(content[AMMO])

                        packs = await self.agent.check_objects_at_step(
                            id_agent, behaviour=self
                        )
                        fov_objects = self.agent.look(id_agent)
                        content = {PACKS: packs, FOV: fov_objects}
                        msg = Message(to=id_agent)
                        msg.set_metadata(PERFORMATIVE, PERFORMATIVE_DATA)
                        msg.body = json.dumps(content)

                        await self.send(msg)
                        if self.agent.check_game_finished(id_agent):
                            await self.agent.inform_game_finished("ALLIED", self)
                            logger.success(
                                "\n\nManager:  GAME FINISHED!! Winner Team: ALLIED! (Target Returned)\n"
                            )
                except Exception as e:
                    logger.warning("Exception at DataFromTroopBehaviour: {}".format(e))
                    logger.warning(traceback.format_exc())

        template = Template()
        template.set_metadata(PERFORMATIVE, PERFORMATIVE_DATA)

        self.add_behaviour(DataFromTroopBehaviour(), template)

    # Behaviour to handle Shot messages
    def launch_shoot_responder_behaviour(self):
        class ShootResponderBehaviour(CyclicBehaviour):
            async def run(self):
                msg = await self.receive(timeout=LONG_RECEIVE_WAIT)
                if msg:
                    content = json.loads(msg.body)
                    shooter_id = content[NAME]
                    aim = int(content[AIM])
                    shots = int(content[SHOTS])
                    victim_pos = Vector3D(x=content[X], y=content[Y], z=content[Z])
                    try:
                        shooter = self.agent.agents[shooter_id]
                    except KeyError:
                        return

                    victim = self.agent.shoot(shooter_id, victim_pos)
                    self.agent.game_statistic.shoot(victim, shooter.team)

                    if victim is None:
                        return

                    damage = 2 if shooter.type == CLASS_SOLDIER else 1
                    damage *= max(0, shots)
                    victim.health -= damage
                    logger.info("Victim hit: {}".format(victim))

                    if victim.health <= 0:
                        victim.health = 0
                        logger.info("Agent {} died.".format(victim.jid))

                        if victim.is_carrying_objective:
                            victim.is_carrying_objective = False
                            logger.info(
                                "Agent {} lost the ObjectivePack.".format(victim.jid)
                            )

                            for din_object in self.agent.din_objects.values():

                                if din_object.type == PACK_OBJPACK:
                                    din_object.is_taken = False
                                    din_object.owner = 0
                                    msg_pack = Message(to=str(din_object.jid))
                                    msg_pack.set_metadata(
                                        PERFORMATIVE, PERFORMATIVE_PACK_LOST
                                    )
                                    din_object.position.x = victim.locate.position.x
                                    din_object.position.y = victim.locate.position.y
                                    din_object.position.z = victim.locate.position.z
                                    msg_pack.body = json.dumps(
                                        {
                                            X: victim.locate.position.x,
                                            Y: victim.locate.position.y,
                                            Z: victim.locate.position.z,
                                        }
                                    )
                                    await self.send(msg_pack)

                                    # Statistics
                                    self.agent.game_statistic.objective_lost(
                                        victim.team
                                    )
                                    break

                    msg_shot = Message(to=victim.jid)
                    msg_shot.set_metadata(PERFORMATIVE, PERFORMATIVE_SHOOT)
                    msg_shot.body = json.dumps({DEC_HEALTH: damage})
                    await self.send(msg_shot)

        template = Template()
        template.set_metadata(PERFORMATIVE, PERFORMATIVE_SHOOT)
        self.add_behaviour(ShootResponderBehaviour(), template)

    # Behaviour to handle Pack Management: Creation and Destruction
    def launch_pack_management_responder_behaviour(self):
        class PackManagementResponderBehaviour(CyclicBehaviour):
            async def run(self):
                msg = await self.receive(LONG_RECEIVE_WAIT)
                if msg:
                    content = json.loads(msg.body)

                    id_ = content[NAME]
                    action = content[ACTION]

                    if action == DESTROY:
                        self.agent.game_statistic.pack_destroyed(
                            self.agent.din_objects[id_]
                        )

                        try:
                            del self.agent.din_objects[id_]
                            logger.info("Pack removed")
                        except KeyError:
                            logger.info("Pack {} cannot be erased".format(id_))
                        return

                    if action == CREATE:
                        type_ = int(content[TYPE])
                        team = int(content[TEAM])

                        x = float(content[X])
                        y = float(content[Y])
                        z = float(content[Z])

                        din_object = DinObject()
                        din_object.jid = msg.sender
                        din_object.type = type_
                        if din_object.type == PACK_OBJPACK:
                            din_object.render_id = 1
                        else:
                            din_object.render_id = abs(hash(din_object.jid)) % 1024
                        din_object.team = team
                        din_object.position.x = x
                        din_object.position.y = y
                        din_object.position.z = z

                        self.agent.din_objects[din_object.jid] = din_object
                        logger.info("Added DinObject {}".format(din_object))

                        self.agent.game_statistic.pack_created(din_object, team)

                    else:
                        logger.warning("Action not identified: {}".format(action))
                        return

        template = Template()
        template.set_metadata(PERFORMATIVE, PERFORMATIVE_PACK)
        self.add_behaviour(PackManagementResponderBehaviour(), template)

    # Behaviour to inform all agents that game has finished by time
    def launch_game_timeout_inform_behaviour(self):
        class GameTimeoutInformBehaviour(TimeoutBehaviour):
            async def run(self):
                logger.success(
                    "\n\nManager:  GAME FINISHED!! Winner Team: AXIS! (Time Expired)\n"
                )
                await self.agent.inform_game_finished("AXIS!", self)

        timeout = datetime.datetime.now() + datetime.timedelta(seconds=self.match_time)
        self.add_behaviour(GameTimeoutInformBehaviour(start_at=timeout))

    # Behaviour to inform all agents that game has finished because all allied troops died
    def launch_check_allied_health(self):
        class CheckAlliedHealthBehaviour(PeriodicBehaviour):
            async def run(self):
                allied_alive = False
                for agent in self.agent.agents.values():
                    if agent.team == TEAM_AXIS:
                        continue
                    if agent.health > 0:
                        allied_alive = True
                        break
                if not allied_alive:
                    logger.success("\n\nManager:  GAME FINISHED!! Winner Team: AXIS!\n")
                    await self.agent.inform_game_finished("AXIS!", self)

        self.add_behaviour(CheckAlliedHealthBehaviour(20 * self.fps))

    async def check_objects_at_step(self, id_agent, behaviour):

        if len(self.din_objects) <= 0:
            return

        if self.agents[id_agent].health <= 0:
            return

        xmin = self.agents[id_agent].locate.position.x - WIDTH
        zmin = self.agents[id_agent].locate.position.z - WIDTH
        xmax = self.agents[id_agent].locate.position.x + WIDTH
        zmax = self.agents[id_agent].locate.position.z + WIDTH

        packs = []

        keys = list(self.din_objects.keys())
        for key in keys:
            if key not in self.din_objects:
                continue
            din_object = self.din_objects[key]
            if (
                din_object.type == PACK_MEDICPACK
                and self.agents[id_agent].health >= 100
            ):
                continue
            if din_object.type == PACK_AMMOPACK and self.agents[id_agent].ammo >= 100:
                continue
            if (
                din_object.type == PACK_OBJPACK
                and din_object.is_taken
                and din_object.owner != 0
            ):
                continue

            if (
                xmin <= din_object.position.x <= xmax
                and zmin <= din_object.position.z <= zmax
            ):
                # Agent has stepped on pack
                id_ = din_object.jid
                type_ = din_object.type
                owner = str(din_object.jid)
                content = None

                team = self.agents[id_agent].team
                self.game_statistic.pack_taken(din_object, team)

                if din_object.type == PACK_MEDICPACK:
                    quantity = DEFAULT_PACK_QTY
                    try:
                        del self.din_objects[id_]
                        logger.info(
                            self.agents[id_agent].jid
                            + ": got a medic pack "
                            + str(din_object.jid)
                        )
                        content = {TYPE: type_, QTY: quantity}

                    except KeyError:
                        logger.error("Could not delete the din object {}".format(id_))

                elif din_object.type == PACK_AMMOPACK:
                    quantity = DEFAULT_PACK_QTY
                    try:
                        del self.din_objects[id_]
                        logger.info(
                            self.agents[id_agent].jid
                            + ": got an ammo pack "
                            + str(din_object.jid)
                        )
                        content = {TYPE: type_, QTY: quantity}
                    except KeyError:
                        logger.error("Could not delete the din object {}".format(id_))

                elif din_object.type == PACK_OBJPACK:
                    if team == TEAM_ALLIED:
                        logger.info(
                            "{}: got the objective pack ".format(
                                self.agents[id_agent].jid
                            )
                        )
                        din_object.is_taken = True
                        din_object.owner = id_agent
                        (
                            din_object.position.x,
                            din_object.position.y,
                            din_object.position.z,
                        ) = (0.0, 0.0, 0.0)
                        self.agents[id_agent].is_carrying_objective = True
                        content = {TYPE: type_, QTY: 0, TEAM: TEAM_ALLIED}

                    elif team == TEAM_AXIS:
                        if din_object.is_taken:
                            logger.info(
                                f"{self.agents[id_agent].jid}: returned the objective pack {din_object.jid}"
                            )
                            din_object.is_taken = False
                            din_object.owner = 0
                            din_object.position.x = self.map.get_target_x()
                            din_object.position.y = self.map.get_target_y()
                            din_object.position.z = self.map.get_target_z()
                            content = {TYPE: type_, QTY: 0, TEAM: TEAM_AXIS}

                # // Send a destroy/taken msg to pack and an inform msg to agent
                if content:
                    content = json.dumps(content)
                    msg = Message(to=owner)
                    msg.set_metadata(PERFORMATIVE, PERFORMATIVE_PACK_TAKEN)
                    msg.body = content
                    await behaviour.send(msg)
                    packs.append(content)
        return packs

    def look(self, name):
        fov_objects = self.get_objects_in_field_of_view(name)
        content = []
        for fov_object in fov_objects:
            obj = {
                TEAM: fov_object.team,
                TYPE: fov_object.type,
                ANGLE: fov_object.angle,
                DISTANCE: fov_object.distance,
                HEALTH: fov_object.health,
                X: fov_object.position.x,
                Y: fov_object.position.y,
                Z: fov_object.position.z,
            }
            content.append(obj)
        return content

    def get_objects_in_field_of_view(self, id_agent):
        objects_in_sight = list()
        agent = None

        # for k, a in self.agents.items():
        #    if a.jid == id_agent:
        #        agent = a
        #        break
        agent = self.agents[id_agent]

        if agent is None or agent.locate.heading.length() == 0:
            return objects_in_sight

        dot_angle = float(agent.locate.angle)

        # am I watching agents?
        for a in self.agents.values():
            if a.jid == id_agent:
                continue
            if (
                a.health <= MIN_HEALTH
            ):  # WARNING, we may be interested in seeing dead agents
                continue

            v = Vector3D(v=a.locate.position)
            v.sub(agent.locate.position)

            distance = v.length()

            # check distance
            # get distance to the closest wall
            distance_terrain = self.intersect_with_walls(
                agent.locate.position, v
            )  # a.locate.heading)

            # check distance
            if distance < agent.locate.view_radius and distance < distance_terrain:

                # check angle
                angle = agent.locate.heading.dot(v)
                try:
                    angle /= agent.locate.heading.length() * v.length()
                except ZeroDivisionError:
                    angle = 0

                if angle >= 0:
                    angle = min(1, angle)
                    angle = math.acos(angle)
                    if angle <= dot_angle:
                        s = Sight()
                        s.distance = distance
                        s.m_id = a.jid
                        s.position = a.locate.position
                        s.team = a.team
                        s.type = a.type
                        s.angle = angle
                        s.health = a.health
                        objects_in_sight.append(s)

        # am I watching objects?
        if len(self.din_objects) > 0:

            for din_object in self.din_objects.values():

                v = Vector3D(v=din_object.position)
                v.sub(agent.locate.position)

                distance = v.length()

                # check distance
                # get distance to the closest wall
                distance_terrain = self.intersect_with_walls(
                    agent.locate.position, v
                )  # a.locate.heading)

                if distance < agent.locate.view_radius and distance < distance_terrain:

                    angle = agent.locate.heading.dot(v)
                    try:
                        angle /= agent.locate.heading.length() * v.length()
                    except ZeroDivisionError:
                        angle = 0

                    if angle >= 0:
                        angle = min(1, angle)
                        angle = math.acos(angle)
                        if angle <= dot_angle:
                            s = Sight()
                            s.distance = distance
                            s.m_id = din_object.jid
                            s.position = din_object.position
                            s.team = din_object.team
                            s.type = din_object.type
                            s.angle = angle
                            s.health = -1
                            objects_in_sight.append(s)

        return objects_in_sight

    def shoot(self, shooter_agent_id, victim_position):
        """
        Agent with id id_agent shoots
        :param shooter_agent_id: agent who shoots
        :param victim_position: the coordinates of the victim to be shot
        :return: agent shot or None
        """
        victim = None
        min_distance = 1e10

        if random.random() <= MISSING_SHOT_PROBABILITY:
            return None

        try:
            shooter_agent = Mobile()
            shooter_agent.position = self.agents[shooter_agent_id].locate.position
            shooter_agent.destination = Vector3D(victim_position)
            shooter_agent.calculate_new_orientation(shooter_agent.destination)
        except KeyError:
            return None

        victim, min_distance = self._check_agents_in_row(
            shooter_agent_id, shooter_agent, victim, min_distance
        )

        if victim is not None:
            shooter_agent.destination = victim.locate.position
            shooter_agent.calculate_new_orientation(shooter_agent.destination)
            distance_to_obstacle = self.intersect_with_walls(
                shooter_agent.position, shooter_agent.heading, min_distance
            )
            if distance_to_obstacle != 0.0 and distance_to_obstacle < min_distance:
                victim = None

        return victim

    def _check_agents_in_row(
        self, shooter_agent_id, shooter_agent, victim, min_distance
    ):
        for agent in self.agents.values():
            if agent.jid == shooter_agent_id:
                continue
            if agent.health <= 0:
                continue

            shooter_position = Vector3D(shooter_agent.position)
            shooter_position.sub(agent.locate.position)

            dv = shooter_position.dot(shooter_agent.heading)
            d2 = shooter_agent.heading.dot(shooter_agent.heading)
            sq = (dv * dv) - ((d2 * shooter_position.dot(shooter_position)) - 4)

            if sq >= 0 and d2 != 0:
                sq = math.sqrt(sq)
                dist1 = (-dv + sq) / d2
                dist2 = (-dv - sq) / d2
                distance = dist1 if dist1 < dist2 else dist2

                if 0 < distance < min_distance:
                    min_distance = distance
                    victim = agent
        return victim, min_distance

    def intersect_with_walls(self, origin, vector, distance=1e10):
        """
        :param origin:
        :param vector:
        :param distance:
        :return: 0.0 if it does not intersect
        """

        try:

            if vector.length() == 0:
                return 0.0

            step = Vector3D(v=vector)
            step.normalize()
            inc = 0
            sgn = 1.0
            e = 0.0

            if abs(step.x) > abs(step.z):

                if step.z < 0:
                    sgn = -1

                step.x /= abs(step.x)
                step.z /= abs(step.x)
            else:

                if step.x < 0:
                    sgn = -1

                inc = 1
                step.x /= abs(step.z)
                step.z /= abs(step.z)

            error = Vector3D(x=0, y=0, z=0)
            point = Vector3D(v=origin)

            while True:

                if inc == 0:

                    if e + abs(step.z) + 0.5 >= 1:
                        point.z += sgn
                        e -= 1

                    e += abs(step.z)
                    point.x += step.x
                else:

                    if e + abs(step.x) + 0.5 >= 1:
                        point.x += sgn
                        e -= 1

                    e += abs(step.x)
                    point.z += step.z

                if not self.map.can_walk(
                    int(math.floor(point.x)), int(math.floor(point.z))
                ):
                    return error.length()

                if point.x < 0 or point.y < 0 or point.z < 0:
                    break
                if point.x >= (self.map.get_size_x()) or point.z >= (
                    self.map.get_size_z()
                ):
                    break
                error.add(step)
                if error.length() > distance:
                    return error.length()
        except Exception as e:
            logger.error(
                "INTERSECT FAILED: (origin: {}) (vector: {}): {}".format(
                    origin, vector, e
                )
            )

        return 0.0

    def check_game_finished(self, id_agent):
        if self.agents[id_agent].team == TEAM_AXIS:
            return False
        if not self.agents[id_agent].is_carrying_objective:
            return False

        if (
            self.map.allied_base.init.x
            < self.agents[id_agent].locate.position.x
            < self.map.allied_base.end.x
            and self.map.allied_base.init.z
            < self.agents[id_agent].locate.position.z
            < self.map.allied_base.end.z
        ):
            return True
        return False

    async def create_objectives(self):

        jid = "objectivepack@" + self.domain
        self.objective_agent = ObjectivePack(
            name=jid,
            passwd="secret",
            manager_jid=str(self.jid),
            x=self.map.get_target_x(),
            z=self.map.get_target_z(),
            team=TEAM_NONE,
        )
        await self.objective_agent.start()

    async def inform_objectives(self, behaviour):

        msg = Message()
        msg.set_metadata(PERFORMATIVE, PERFORMATIVE_OBJECTIVE)
        content = {
            X: self.map.get_target_x(),
            Y: self.map.get_target_y(),
            Z: self.map.get_target_z(),
        }
        msg.body = json.dumps(content)
        for agent in self.agents.values():
            msg.to = agent.jid
            logger.info("Sending objective to {}: {}".format(agent.jid, msg))
            await behaviour.send(msg)
        logger.info("Manager: Sending Objective notification to agents")

    async def inform_game_finished(self, winner_team, behaviour):

        for agent in self.agents.values():
            msg = Message()
            msg.set_metadata(PERFORMATIVE, PERFORMATIVE_GAME)
            msg.body = "GAME FINISHED!! Winner Team: " + str(winner_team)
            msg.to = agent.jid
            await behaviour.send(msg)
        for st in self.render_server.get_connections():
            try:
                # st.send_msg_to_render_engine(TCP_COM, "FINISH " + " GAME FINISHED!! Winner Team: " + str(winner_team))
                st.send_msg_to_render_engine(TCP_COM, QUIT_MSG)
            except:
                pass

        self.print_statistics(winner_team)

        await self.stop()

    def print_statistics(self, winner_team):

        allied_alive_players = 0
        axis_alive_players = 0
        allied_health = 0
        axis_health = 0

        self.game_statistic.match_duration = time.time() - self.match_init
        logger.info("Match took {} seconds".format(self.game_statistic.match_duration))

        for agent in self.agents.values():
            if agent.team == TEAM_ALLIED:
                allied_health += agent.health
                if agent.health > 0:
                    allied_alive_players = allied_alive_players + 1
            else:
                axis_health += agent.health
                if agent.health > 0:
                    axis_alive_players = axis_alive_players + 1

        self.game_statistic.calculate_data(
            allied_alive_players, axis_alive_players, allied_health, axis_health
        )

        try:
            fw = open("pygomas_stats.txt", "w+")

            fw.write(self.game_statistic.dumps(winner_team))

            fw.close()

        except Exception as e:
            logger.exception("COULD NOT WRITE STATISTICS TO FILE: {}".format(e))



# Ontologies
ACTION: str = "ACTION"
AIM: str = "aim"
ANGLE: str = "angle"
CREATE: str = "CREATE"
DEC_AMMO: str = "dec_ammo"
DEC_HEALTH: str = "dec_health"
DESTROY: str = "DESTROY"
DISTANCE: str = "distance"
FOV: str = "fov"
HEAD_X: str = "headx"
HEAD_Y: str = "heady"
HEAD_Z: str = "headz"
MAP: str = "map"
PACKS: str = "PACKS"
QTY: str = "qty"
SHOTS: str = "shots"
VEL_X: str = "xvel"
TYPE: str = "type"
VEL_Y: str = "yvel"
VEL_Z: str = "zvel"
X: str = "x"
Y: str = "y"
Z: str = "z"

# Performatives
PERFORMATIVE: str = "performative"
PERFORMATIVE_BDI: str = "BDI"
PERFORMATIVE_CFA: str = "cfa"
PERFORMATIVE_CFB: str = "cfb"
PERFORMATIVE_CFM: str = "cfm"
PERFORMATIVE_DATA: str = "data"
PERFORMATIVE_DEREGISTER_AGENT: str = "deregister_agent"
PERFORMATIVE_DEREGISTER_SERVICE: str = "deregister_service"
PERFORMATIVE_GAME: str = "game"
PERFORMATIVE_GET: str = "get"
PERFORMATIVE_INFORM: str = "inform"
PERFORMATIVE_INIT: str = "init"
PERFORMATIVE_MOVE: str = "move"
PERFORMATIVE_OBJECTIVE: str = "objective"
PERFORMATIVE_PACK: str = "pack"
PERFORMATIVE_PACK_LOST: str = "pack_lost"
PERFORMATIVE_REGISTER_AGENT: str = "register_agent"
PERFORMATIVE_REGISTER_SERVICE: str = "register"
PERFORMATIVE_SERVICES: str = "services"
PERFORMATIVE_SIGHT: str = "sight"
PERFORMATIVE_SHOOT: str = "shot"

# Services
AMMO_SERVICE: str = "fieldops"
BACKUP_SERVICE: str = "backup"
MANAGEMENT_SERVICE: str = "management"
MEDIC_SERVICE: str = "medic"

# Set of beliefs
AMMO: str = "ammo"
BASE: str = "base"
CLASS: str = "class"
DESTINATION: str = "destination"
ENEMIES_IN_FOV: str = "enemies_in_fov"
FRIENDS_IN_FOV: str = "friends_in_fov"
FLAG: str = "flag"
HEADING: str = "heading"
HEALTH: str = "health"
NAME: str = "name"
MY_MEDICS: str = "myMedics"
MY_FIELDOPS: str = "myFieldops"
MY_BACKUPS: str = "myBackups"
PACKS_IN_FOV: str = "packs_in_fov"
PERFORMATIVE_PACK_TAKEN: str = "pack_taken"
PERFORMATIVE_TARGET_REACHED: str = "target_reached"
PERFORMATIVE_FLAG_TAKEN: str = "flag_taken"
POSITION: str = "position"
TEAM: str = "team"
THRESHOLD_HEALTH: str = "threshold_health"
THRESHOLD_AMMO: str = "threshold_ammo"
THRESHOLD_AIM: str = "threshold_aim"
THRESHOLD_SHOTS: str = "threshold_shots"
VELOCITY: str = "velocity"



import heapq
import math
import time

import numpy as np
from loguru import logger

OBSTACLE = 0
MANHATTAN = 1
EUCLIDEAN = 2


class JPSAlgorithm(object):
    """Jump-Point-Search algorithm"""

    def __init__(self, map_array, downsample=False):
        self.array = map_array
        self.downsample = downsample
        if downsample:
            m, n = self.array.shape
            self.array = self.array.reshape(m // 8, 8, n // 8, 8).min(axis=(1, 3))

    @staticmethod
    def heuristic(a, b, hchoice=EUCLIDEAN):
        if hchoice == MANHATTAN:
            xdist = math.fabs(b[0] - a[0])
            ydist = math.fabs(b[1] - a[1])
            if xdist > ydist:
                return 14 * ydist + 10 * (xdist - ydist)
            else:
                return 14 * xdist + 10 * (ydist - xdist)
        if hchoice == EUCLIDEAN:
            return math.sqrt((b[0] - a[0]) ** 2 + (b[1] - a[1]) ** 2)

    def blocked(self, cx, cy, dx, dy):
        if cx + dx < 0 or cx + dx >= self.array.shape[0]:
            return True
        if cy + dy < 0 or cy + dy >= self.array.shape[1]:
            return True
        if dx != 0 and dy != 0:
            if (
                self.array[cx + dx][cy] == OBSTACLE
                and self.array[cx][cy + dy] == OBSTACLE
            ):
                return True
            if self.array[cx + dx][cy + dy] == OBSTACLE:
                return True
        else:
            if dx != 0:
                if self.array[cx + dx][cy] == OBSTACLE:
                    return True
            else:
                if self.array[cx][cy + dy] == OBSTACLE:
                    return True
        return False

    def dblock(self, cx, cy, dx, dy):
        if self.array[cx - dx][cy] == OBSTACLE and self.array[cx][cy - dy] == OBSTACLE:
            return True
        else:
            return False

    @staticmethod
    def direction(cx, cy, px, py):
        dx = int(math.copysign(1, cx - px))
        dy = int(math.copysign(1, cy - py))
        if cx - px == 0:
            dx = 0
        if cy - py == 0:
            dy = 0
        return dx, dy

    def node_neighbours(self, cx, cy, parent):
        neighbours = []
        if type(parent) != tuple:
            for i, j in [
                (-1, 0),
                (0, -1),
                (1, 0),
                (0, 1),
                (-1, -1),
                (-1, 1),
                (1, -1),
                (1, 1),
            ]:
                if not self.blocked(cx, cy, i, j):
                    neighbours.append((cx + i, cy + j))

            return neighbours
        dx, dy = self.direction(cx, cy, parent[0], parent[1])

        if dx != 0 and dy != 0:
            if not self.blocked(cx, cy, 0, dy):
                neighbours.append((cx, cy + dy))
            if not self.blocked(cx, cy, dx, 0):
                neighbours.append((cx + dx, cy))
            if (
                not self.blocked(cx, cy, 0, dy) or not self.blocked(cx, cy, dx, 0)
            ) and not self.blocked(cx, cy, dx, dy):
                neighbours.append((cx + dx, cy + dy))
            if self.blocked(cx, cy, -dx, 0) and not self.blocked(cx, cy, 0, dy):
                neighbours.append((cx - dx, cy + dy))
            if self.blocked(cx, cy, 0, -dy) and not self.blocked(cx, cy, dx, 0):
                neighbours.append((cx + dx, cy - dy))

        else:
            if dx == 0:
                if not self.blocked(cx, cy, dx, 0):
                    if not self.blocked(cx, cy, 0, dy):
                        neighbours.append((cx, cy + dy))
                    if self.blocked(cx, cy, 1, 0):
                        neighbours.append((cx + 1, cy + dy))
                    if self.blocked(cx, cy, -1, 0):
                        neighbours.append((cx - 1, cy + dy))

            else:
                if not self.blocked(cx, cy, dx, 0):
                    if not self.blocked(cx, cy, dx, 0):
                        neighbours.append((cx + dx, cy))
                    if self.blocked(cx, cy, 0, 1):
                        neighbours.append((cx + dx, cy + 1))
                    if self.blocked(cx, cy, 0, -1):
                        neighbours.append((cx + dx, cy - 1))
        return neighbours

    def jump(self, cx, cy, dx, dy, goal):

        nx = cx + dx
        ny = cy + dy
        if self.blocked(nx, ny, 0, 0):
            return None

        if (nx, ny) == goal:
            return nx, ny

        ox = nx
        oy = ny

        if dx != 0 and dy != 0:
            while True:
                if (
                    not self.blocked(ox, oy, -dx, dy)
                    and self.blocked(ox, oy, -dx, 0)
                    or not self.blocked(ox, oy, dx, -dy)
                    and self.blocked(ox, oy, 0, -dy)
                ):
                    return ox, oy

                if (
                    self.jump(ox, oy, dx, 0, goal) is not None
                    or self.jump(ox, oy, 0, dy, goal) is not None
                ):
                    return ox, oy

                ox += dx
                oy += dy

                if self.blocked(ox, oy, 0, 0):
                    return None

                if self.dblock(ox, oy, dx, dy):
                    return None

                if (ox, oy) == goal:
                    return ox, oy
        else:
            if dx != 0:
                while True:
                    if (
                        not self.blocked(ox, ny, dx, 1)
                        and self.blocked(ox, ny, 0, 1)
                        or not self.blocked(ox, ny, dx, -1)
                        and self.blocked(ox, ny, 0, -1)
                    ):
                        return ox, ny

                    ox += dx

                    if self.blocked(ox, ny, 0, 0):
                        return None

                    if (ox, ny) == goal:
                        return ox, ny

            else:
                while True:
                    if (
                        not self.blocked(nx, oy, 1, dy)
                        and self.blocked(nx, oy, 1, 0)
                        or not self.blocked(nx, oy, -1, dy)
                        and self.blocked(nx, oy, -1, 0)
                    ):
                        return nx, oy

                    oy += dy

                    if self.blocked(nx, oy, 0, 0):
                        return None

                    if (nx, oy) == goal:
                        return nx, oy

        return self.jump(nx, ny, dx, dy, goal)

    def identify_successors(self, cx, cy, came_from, goal):
        successors = []
        neighbours = self.node_neighbours(cx, cy, came_from.get((cx, cy), 0))

        for cell in neighbours:
            dx = cell[0] - cx
            dy = cell[1] - cy

            jump_point = self.jump(cx, cy, dx, dy, goal)

            if jump_point is not None:
                successors.append(jump_point)

        return successors

    def get_path(self, start, goal, hchoice=EUCLIDEAN):

        start = np.round(start)
        goal = np.round(goal)
        start = (int(start[0]), int(start[1]))
        goal = (int(goal[0]), int(goal[1]))
        logger.debug("Finding path from {} to {}".format(start, goal))
        start_time = time.time()

        came_from = {}
        close_set = set()
        gscore = {start: 0}
        fscore = {start: self.heuristic(start, goal, hchoice)}

        pqueue = []

        heapq.heappush(pqueue, (fscore[start], start))

        while pqueue:

            current = heapq.heappop(pqueue)[1]
            if current == goal:
                data = []
                while current in came_from:
                    data.append(current)
                    current = came_from[current]
                data.append(start)
                data = data[::-1]
                logger.debug(
                    "Got path from {} to {} with score {} in {:.02f}".format(
                        start, goal, gscore[goal], time.time() - start_time
                    )
                )
                return data

            close_set.add(current)

            successors = self.identify_successors(
                current[0], current[1], came_from, goal
            )

            for successor in successors:
                jump_point = successor

                if (
                    jump_point in close_set
                ):  # and tentative_g_score >= gscore.get(jump_point,0):
                    continue

                tentative_g_score = gscore[current] + self.lenght(
                    current, jump_point, hchoice
                )

                if tentative_g_score < gscore.get(jump_point, 0) or jump_point not in [
                    j[1] for j in pqueue
                ]:
                    came_from[jump_point] = current
                    gscore[jump_point] = tentative_g_score
                    fscore[jump_point] = tentative_g_score + self.heuristic(
                        jump_point, goal, hchoice
                    )
                    heapq.heappush(pqueue, (fscore[jump_point], jump_point))
        return False

    def lenght(self, current, jumppoint, hchoice):
        dx, dy = self.direction(current[0], current[1], jumppoint[0], jumppoint[1])
        dx = math.fabs(dx)
        dy = math.fabs(dy)
        lx = math.fabs(current[0] - jumppoint[0])
        ly = math.fabs(current[1] - jumppoint[1])
        if hchoice == MANHATTAN:
            if dx != 0 and dy != 0:
                lenght = lx * 14
                return lenght
            else:
                lenght = (dx * lx + dy * ly) * 10
                return lenght
        if hchoice == EUCLIDEAN:
            return math.sqrt(
                (current[0] - jumppoint[0]) ** 2 + (current[1] - jumppoint[1]) ** 2
            )



#!/usr/bin/env python
# -*- coding: UTF8 -*-
import argparse
import socket
import sys
import os
import traceback
import copy
import pygame
from pygame import gfxdraw
import math

objective_x = -1
objective_y = -1
allied_base = None
axis_base = None
graph = {}
agents = {}
dins = {}
factor = 2
screen = None
font = None
maps_path = None

iteration = 0


tile_size = 24
horizontal_tiles = 32
vertical_tiles = 32

map_width = tile_size * horizontal_tiles
map_height = tile_size * vertical_tiles

xdesp = 0
ydesp = 0


def agl_parse(data):
    global allied_base
    global axis_base
    global objective_x
    global objective_y
    global agents
    global dins

    dins = {}
    agl = data.split()
    nagents = int(agl[1])
    agl = agl[2:]
    separator = nagents * 15
    agent_data = agl[:separator]
    din_data = agl[separator:]
    for i in range(nagents):
        agents[agent_data[0]] = {
            "type": agent_data[1],
            "team": agent_data[2],
            "health": agent_data[3],
            "ammo": agent_data[4],
            "carrying": agent_data[5],
            "posx": agent_data[6].strip("(,)"),
            "posy": agent_data[7].strip("(,)"),
            "posz": agent_data[8].strip("(,)"),
            "angx": agent_data[12].strip("(,)"),
            "angy": agent_data[13].strip("(,)"),
            "angz": agent_data[14].strip("(,)"),
        }
        agent_data = agent_data[15:]

    ndin = int(din_data[0])
    din_data = din_data[1:]
    for din in range(ndin):
        dins[din_data[0]] = {
            "type": din_data[1],
            "posx": din_data[2].strip("(,)"),
            "posy": din_data[3].strip("(,)"),
            "posz": din_data[4].strip("(,)"),
        }
        din_data = din_data[5:]


def draw2():

    global agents
    global factor
    global xdesp
    global ydesp
    global tile_size
    global iteration
    global screen
    global font

    events = pygame.event.get()
    for event in events:
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_LEFT:
                xdesp += tile_size
            if event.key == pygame.K_RIGHT:
                xdesp -= tile_size
            if event.key == pygame.K_DOWN:
                ydesp -= tile_size
            if event.key == pygame.K_UP:
                ydesp += tile_size
            if event.key == pygame.K_x:
                tile_size += 2
            if event.key == pygame.K_z:
                tile_size -= 2

    # Clear screen
    color_background = (0, 0, 0)
    pygame.draw.rect(screen, color_background, (0, 0, map_width, map_height))

    # Draw Map
    color_wall = (100, 100, 100)
    for y in range(0, len(list(graph.items()))):
        for x in range(0, 32):
            try:
                if list(graph.items())[y][1][x] == "*":
                    pygame.draw.rect(
                        screen,
                        color_wall,
                        (
                            x * tile_size + xdesp,
                            y * tile_size + ydesp,
                            tile_size,
                            tile_size,
                        ),
                    )
            except:
                pass

    # Draw bases
    if allied_base is not None:
        color = (255, 0, 0)
        xpos = int(allied_base[0]) * tile_size + xdesp
        ypos = int(allied_base[1]) * tile_size + ydesp
        xwidth = int(allied_base[2]) * tile_size - xpos + tile_size + xdesp
        ywidth = int(allied_base[3]) * tile_size - ypos + tile_size + ydesp

        pygame.draw.rect(screen, color, (xpos, ypos, xwidth, ywidth))

    if axis_base is not None:
        color = (0, 0, 255)
        xpos = int(axis_base[0]) * tile_size + xdesp
        ypos = int(axis_base[1]) * tile_size + ydesp
        xwidth = int(axis_base[2]) * tile_size - xpos + tile_size + xdesp
        ywidth = int(axis_base[3]) * tile_size - ypos + tile_size + ydesp

        pygame.draw.rect(screen, color, (xpos, ypos, xwidth, ywidth))

    # Draw items
    for i in range(0, len(list(dins.items()))):
        posx = int(float(list(dins.items())[i][1]["posx"]) * (tile_size / 8.0)) + xdesp
        posy = int(float(list(dins.items())[i][1]["posz"]) * (tile_size / 8.0)) + ydesp

        item_type = {"1001": "M", "1002": "A", "1003": "F"}.get(
            list(dins.items())[i][1]["type"], "X"
        )

        color = {
            "1001": (255, 255, 255),
            "1002": (255, 255, 255),
            "1003": (255, 255, 0),
        }.get(list(dins.items())[i][1]["type"], "X")

        pygame.draw.circle(screen, color, [posx, posy], 6)
        text = font.render(item_type, True, (0, 0, 0))
        screen.blit(text, (posx - text.get_width() // 2, posy - text.get_height() // 2))

    # Draw units
    for i in list(agents.items()):
        health = float(i[1]["health"])

        if float(health) > 0:

            carrying = i[1]["carrying"]

            agent_type = {"0": "X", "1": "*", "2": "+", "3": "Y", "4": "^"}.get(
                i[1]["type"], "X"
            )

            team = {"100": (255, 100, 100), "200": (100, 100, 255)}.get(
                i[1]["team"], (255, 255, 0)
            )

            team_aplha = {"100": (255, 100, 100, 100), "200": (100, 100, 255, 100)}.get(
                i[1]["team"], (255, 255, 0, 255)
            )

            ammo = float(i[1]["ammo"])

            posx = int(float(i[1]["posx"]) * tile_size / 8.0) + xdesp
            posy = int(float(i[1]["posz"]) * tile_size / 8.0) + ydesp

            # calcula direccion
            angx = float(i[1]["angx"])
            angy = float(i[1]["angz"])

            if angx == 0:
                div = 1000
            else:
                div = angy / angx

            if angy >= 0 and angx >= 0:  # q1
                angle = math.atan(div) * (180 / math.pi)
            elif angy >= 0 and angx <= 0:  # q2
                angle = math.atan(div) * (180 / math.pi) + 180
            elif angy <= 0 and angx <= 0:  # q3
                angle = math.atan(div) * (180 / math.pi) + 180
            else:  # q4
                angle = math.atan(div) * (180 / math.pi) + 360

            # imprime ficha
            pygame.draw.circle(screen, team, [posx, posy], 8)
            # imprime identificador
            text = font.render(i[0], True, (255, 255, 255))
            screen.blit(
                text,
                (posx - text.get_width() // 2 + 15, posy - text.get_height() // 2 - 15),
            )
            # imprime vida
            pygame.gfxdraw.aacircle(screen, posx, posy, 10, (255, 0, 0))
            pygame.gfxdraw.aacircle(screen, posx, posy, 9, (255, 0, 0))
            pygame.gfxdraw.arc(
                screen, posx, posy, 10, 0, int(health * 3.6) - 1, (0, 255, 0)
            )
            pygame.gfxdraw.arc(
                screen, posx, posy, 9, 0, int(health * 3.6) - 1, (0, 255, 0)
            )
            # imprime municion
            if ammo >= 1:
                pygame.gfxdraw.arc(
                    screen, posx, posy, 6, 0, int(ammo * 3.6) - 1, (255, 255, 255)
                )
                pygame.gfxdraw.arc(
                    screen, posx, posy, 7, 0, int(ammo * 3.6) - 1, (255, 255, 255)
                )

            # lleva la bandera
            if carrying == "1":
                pygame.draw.circle(screen, (255, 255, 0), [posx, posy], 5)

            # imprime cono de vision
            for j in range(0, int(48 * (tile_size / 8)), 1):
                pygame.gfxdraw.arc(
                    screen, posx, posy, j, int(-45 + angle), int(45 + angle), team_aplha
                )

            # imprime funcion
            text = font.render(agent_type, True, (0, 0, 0))
            screen.blit(
                text, (posx - text.get_width() // 2, posy - text.get_height() // 2)
            )

    pygame.display.flip()
    iteration += 1

    agents_json = {}
    for i in list(agents.items()):
        agents_json[i[0]] = i[1]

    items_json = {}
    for i in list(dins.items()):
        items_json[i[0]] = i[1]

    json_data = {"agents": agents_json, "items": items_json}


def loadMap(map_name):
    global allied_base
    global axis_base
    global objective_x
    global objective_y
    global maps_path

    if maps_path is not None:
        path = f"{maps_path}{os.sep}{map_name}{os.sep}{map_name}"
    else:
        this_dir, _ = os.path.split(__file__)
        path = f"{this_dir}{os.sep}maps{os.sep}{map_name}{os.sep}{map_name}"

    mapf = open(f"{path}.txt", "r")
    cost = open(f"{path}_cost.txt", "r")

    for line in mapf.readlines():
        if "pGomas_OBJECTIVE" in line:
            l = line.split()
            objective_x = copy.copy(int(l[1]))
            objective_y = copy.copy(int(l[2]))
        elif "pGomas_SPAWN_ALLIED" in line:
            l = line.split()
            l.pop(0)
            allied_base = copy.copy(l)
        elif "pGomas_SPAWN_AXIS" in line:
            l = line.split()
            l.pop(0)
            axis_base = copy.copy(l)
    mapf.close()

    y = 0
    for line in cost.readlines():
        graph[y] = line.strip("\r\n")
        y += 1
    cost.close()


def main(address="localhost", port=8001, maps=None):
    global f
    global screen
    global font
    global maps_path

    # Main
    maps_path = maps

    # Init pygame
    pygame.init()
    font = pygame.font.SysFont("ttf-font-awesome", 12)

    # Set the height and width of the screen
    size = [map_width, map_height]
    screen = pygame.display.set_mode(size)

    try:
        # Init socket
        s = None
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if s:
            s.connect((address, port))
            rfile = s.makefile("r", -1)
            wfile = s.makefile("w", 20)
            data = rfile.readline()

            wfile.write("READY\n")
            wfile.close()
            in_loop = True
            while in_loop:
                data = ""
                data = rfile.readline()
                if "COM" in data[0:5]:
                    if "Accepted" in data:
                        pass
                    elif "Closed" in data:
                        in_loop = False
                elif "MAP" in data[0:5]:
                    p = data.split()
                    mapname = p[2]
                    loadMap(mapname)
                elif "AGL" in data[0:5]:
                    agl_parse(data)
                elif "TIM" in data[0:5]:
                    pass
                elif "ERR" in data[0:5]:
                    pass
                else:
                    # Unknown message type
                    pass
                draw2()

            # Close socket
            del rfile
            del wfile
            s.close()

    except Exception as e:
        print("Exception", str(e))
        print("-" * 60)
        traceback.print_exc(file=sys.stdout)
        print("-" * 60)

    finally:
        pygame.quit()
        f.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ip", default="localhost", help="Manager's address to connect the render"
    )
    parser.add_argument(
        "--port", default=8001, help="Manager's port to connect the render"
    )
    parser.add_argument(
        "--maps", default=None, help="The path to your custom maps directory"
    )

    args = parser.parse_args()
    main(args.ip, args.port, args.maps)
    sys.exit(0)



import unittest
from jgomas.CThreshold import CThreshold


class TestCThreshold(unittest.TestCase):

    def test_getters(self):

        thresh = CThreshold()

        health = thresh.GetHealth()
        ammo = thresh.GetAmmo()
        aim = thresh.GetAim()
        shot = thresh.GetShot()
        look = thresh.GetLook()

        self.assertEqual(
            health,
            50
        )

        self.assertEqual(
            ammo,
            50
        )

        self.assertEqual(
            aim,
            1
        )

        self.assertEqual(
            shot,
            1
        )

        self.assertEqual(
            look,
            1
        )

    def test_upper_setters(self):

        thresh = CThreshold()

        health = 110
        ammo = 120
        aim = 30
        shot = 40
        look = 130

        thresh.SetHealth(health)
        thresh.SetAmmo(ammo)
        thresh.SetAim(aim)
        thresh.SetShot(shot)
        thresh.SetLook(look)

        self.assertEqual(
            thresh.m_iHealth,
            100
        )

        self.assertEqual(
            thresh.m_iAmmo,
            100
        )

        self.assertEqual(
            thresh.m_iAim,
            20
        )

        self.assertEqual(
            thresh.m_iShot,
            20
        )

        self.assertEqual(
            thresh.m_iLook,
            100
        )

    def test_lower_setters(self):

        thresh = CThreshold()

        health = -10
        ammo = -20
        aim = 0
        shot = -1
        look = -2

        thresh.SetHealth(health)
        thresh.SetAmmo(ammo)
        thresh.SetAim(aim)
        thresh.SetShot(shot)
        thresh.SetLook(look)

        self.assertEqual(
            thresh.m_iHealth,
            0
        )

        self.assertEqual(
            thresh.m_iAmmo,
            0
        )

        self.assertEqual(
            thresh.m_iAim,
            1
        )

        self.assertEqual(
            thresh.m_iShot,
            1
        )

        self.assertEqual(
            thresh.m_iLook,
            0
        )



import os

this_dir, _ = os.path.split(__file__)
DEFAULT_DATA_PATH = f"{this_dir}{os.sep}maps{os.sep}"


class Config(object):
    def __init__(self, data_path=None):
        self.data_path = data_path if data_path else DEFAULT_DATA_PATH
        if not self.data_path.endswith(os.sep):
            self.data_path += os.sep

    def set_data_path(self, data_path):
        self.data_path = data_path
        if not self.data_path.endswith(os.sep):
            self.data_path += os.sep


MIN_POWER: int = 0
MAX_POWER: int = 100
POWER_UNIT: int = 25
MIN_STAMINA: int = 0
MAX_STAMINA: int = 100
STAMINA_UNIT: int = 5
MIN_AMMO: int = 0
MAX_AMMO: int = 100
MIN_HEALTH: int = 0
MAX_HEALTH: int = 100
TEAM_NONE: int = 0
TEAM_ALLIED: int = 100
TEAM_AXIS: int = 200
PRECISION_Z: float = 0.5
PRECISION_X: float = 0.5
MISSING_SHOT_PROBABILITY: float = 0.1



import unittest
from jgomas.CJGomasAgent import CJGomasAgent
from jgomas.CRegistry import CRegistry


class TestCJGomasAgent(unittest.TestCase):

    def test_add_service(self):
        agent = CJGomasAgent("agent@localhost", "agent", 0)
        agent.start()
        agent.stop()





from .bditroop import BDITroop, CLASS_SOLDIER
from .ontology import BACKUP_SERVICE


class BDISoldier(BDITroop):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.services.append(BACKUP_SERVICE)
        self.eclass = CLASS_SOLDIER



import pygomas.vector
import pygomas.sight
import pygomas.mobile
import pygomas.service
import pygomas.threshold
import pygomas.config
import pygomas.map
import pygomas.agent
import pygomas.server
import pygomas.pack
import pygomas.ammopack
import pygomas.medicpack
import pygomas.objpack

__version__ = "0.4.4"



import random

from loguru import logger
from spade.behaviour import OneShotBehaviour

from .config import POWER_UNIT
from .ammopack import AmmoPack
from .bditroop import BDITroop, CLASS_FIELDOPS
from .ontology import AMMO_SERVICE


class BDIFieldOp(BDITroop):
    packs_delivered = 0
    ammo_pack_offset = 5

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.services.append(AMMO_SERVICE)
        self.eclass = CLASS_FIELDOPS

    def add_custom_actions(self, actions):
        super().add_custom_actions(actions)

        @actions.add(".reload", 0)
        def _cure(agent, term, intention):
            class CreateAmmoPackBehaviour(OneShotBehaviour):
                async def run(self):
                    await self.agent.create_ammo_pack()

            b = CreateAmmoPackBehaviour()
            self.add_behaviour(b)
            yield

    def perform_ammo_action(self):
        # We can give ammo paks if we have power enough...
        if self.get_power() >= POWER_UNIT:
            self.use_power()
            return True
        return False

    async def create_ammo_pack(self):
        """
        Creates ammo packs if possible.

        This method allows to create ammo packs if there is enough power in the agent's power bar.

        :returns number of ammo packs created
        """

        logger.info("{} Creating ammo packs.".format(self.name))
        while self.perform_ammo_action():
            BDIFieldOp.packs_delivered += 1
            name = "ammopack_{}_{}@{}".format(
                self.jid.localpart, BDIFieldOp.packs_delivered, self.jid.domain
            )
            x = self.movement.position.x + random.random() * BDIFieldOp.ammo_pack_offset
            z = self.movement.position.z + random.random() * BDIFieldOp.ammo_pack_offset

            while not self.check_static_position(x, z):
                x = (
                    self.movement.position.x
                    + random.random() * BDIFieldOp.ammo_pack_offset
                )
                z = (
                    self.movement.position.z
                    + random.random() * BDIFieldOp.ammo_pack_offset
                )

            try:
                pack = AmmoPack(
                    name=name, passwd="secret", x=x, z=z, manager_jid=self.manager
                )
                await pack.start()
            except Exception as e:
                logger.warning(
                    "FieldOps {} could not create AmmoPack: {}".format(self.name, e)
                )

            logger.info("AmmoPack {} created.".format(name))



import random
import math
from .vector import Vector3D

MAP_SCALE = 1


class Mobile(object):
    def __init__(self, velocity_value=2):
        self.position = Vector3D()
        self.destination = Vector3D()
        self.velocity = Vector3D()
        self.heading = Vector3D()
        self.view_radius = 50.0
        self.angle = 1.0
        self.min_x = self.min_z = self.max_x = self.max_z = 0
        self.velocity_value = velocity_value

    def set_size(self, max_x, max_z):
        self.min_x = MAP_SCALE
        self.min_z = MAP_SCALE
        self.max_x = (max_x * MAP_SCALE) - MAP_SCALE
        self.max_z = (max_z * MAP_SCALE) - MAP_SCALE

    def calculate_position(self, dt):
        position = Vector3D()
        position.x = self.position.x + (self.velocity.x * dt)
        position.y = self.position.y + (self.velocity.y * dt)  # + (0.5f * t2)
        position.z = self.position.z + (self.velocity.z * dt)
        return position

    def calculate_new_orientation(self, destination):
        dx = float(destination.x - self.position.x)
        dy = float(destination.y - self.position.y)
        dz = float(destination.z - self.position.z)
        f2_norma = math.sqrt((dx * dx + dy * dy + dz * dz))

        if f2_norma > 0:
            self.velocity.x = float(dx / f2_norma)
            self.velocity.y = float(dy / f2_norma)
            self.velocity.z = float(dz / f2_norma)
        else:
            self.velocity.x = 0.0
            self.velocity.y = 0.0
            self.velocity.z = 0.0

        if self.velocity.length() > 0.0001:
            self.heading.x = self.velocity.x
            self.heading.y = self.velocity.y
            self.heading.z = self.velocity.z

        self.velocity.x *= self.velocity_value
        self.velocity.y *= self.velocity_value
        self.velocity.z *= self.velocity_value

    def calculate_new_destination(self, radius_x, radius_y):
        x = self.position.x + ((random.random() * (radius_x * 2)) - radius_x)
        z = self.position.z + ((random.random() * (radius_y * 2)) - radius_y)

        x = min(x, self.min_x)
        x = max(x, self.max_x)
        z = min(z, self.min_z)
        z = max(z, self.max_z)

        return Vector3D(x=x, y=0.0, z=z)

    def get_destination(self):
        return self.destination

    def set_destination(self, destination):
        self.destination = destination

    def get_position(self):
        return self.position

    def get_angle(self):
        return self.angle

    def get_view_radius(self):
        return self.view_radius

    def get_heading(self):
        return self.heading

    def get_velocity(self):
        return self.velocity
